# Copyright 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import testtools
from testtools import matchers as tt_matchers

from keystoneauth1.tests.unit import matchers as ks_matchers

# NOTE(jamielennox): The tests in this file are copied form the non-public
# testtools.tests.matchers.helpers.TestMatchersInterface.


class TestXMLEquals(testtools.TestCase):
    matches_xml = b"""<?xml version="1.0" encoding="UTF-8"?>
<test xmlns="https://docs.openstack.org/identity/api/v2.0">
    <first z="0" y="1" x="2"/>
    <second a="a" b="b"></second>
</test>
"""
    equivalent_xml = b"""<?xml version="1.0" encoding="UTF-8"?>
<test xmlns="https://docs.openstack.org/identity/api/v2.0">
    <second a="a" b="b"/>
    <first z="0" y="1" x="2"></first>
</test>
"""
    mismatches_xml = b"""<?xml version="1.0" encoding="UTF-8"?>
<test xmlns="https://docs.openstack.org/identity/api/v2.0">
    <nope_it_fails/>
</test>
"""
    mismatches_description = """expected =
<test xmlns="https://docs.openstack.org/identity/api/v2.0">
  <first z="0" y="1" x="2"/>
  <second a="a" b="b"/>
</test>

actual =
<test xmlns="https://docs.openstack.org/identity/api/v2.0">
  <nope_it_fails/>
</test>
"""

    matches_matcher = ks_matchers.XMLEquals(matches_xml)
    matches_matches = [matches_xml, equivalent_xml]
    matches_mismatches = [mismatches_xml]
    describe_examples = [
        (mismatches_description, mismatches_xml, matches_matcher)
    ]
    str_examples = [(f'XMLEquals({matches_xml!r})', matches_matcher)]

    def test_matches_match(self):
        matcher = self.matches_matcher
        matches = self.matches_matches
        mismatches = self.matches_mismatches
        for candidate in matches:
            self.assertIsNone(matcher.match(candidate))
        for candidate in mismatches:
            mismatch = matcher.match(candidate)
            self.assertIsNotNone(mismatch)
            self.assertIsNotNone(getattr(mismatch, 'describe', None))

    def test__str__(self):
        # [(expected, object to __str__)].
        examples = self.str_examples
        for expected, matcher in examples:
            self.assertThat(matcher, tt_matchers.DocTestMatches(expected))

    def test_describe_difference(self):
        # [(expected, matchee, matcher), ...]
        examples = self.describe_examples
        for difference, matchee, matcher in examples:
            mismatch = matcher.match(matchee)
            self.assertEqual(difference, mismatch.describe())

    def test_mismatch_details(self):
        # The mismatch object must provide get_details, which must return a
        # dictionary mapping names to Content objects.
        examples = self.describe_examples
        for difference, matchee, matcher in examples:
            mismatch = matcher.match(matchee)
            details = mismatch.get_details()
            self.assertEqual(dict(details), details)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock
import uuid

from oslo_config import cfg
from oslo_config import fixture as config
import stevedore

from keystoneauth1 import exceptions
from keystoneauth1 import loading
from keystoneauth1.loading._plugins.identity import v2
from keystoneauth1.loading._plugins.identity import v3
from keystoneauth1.tests.unit.loading import utils


class ConfTests(utils.TestCase):
    def setUp(self):
        super().setUp()
        self.conf_fixture = self.useFixture(config.Config())

        # NOTE(jamielennox): we register the basic config options first because
        # we need them in place before we can stub them. We will need to run
        # the register again after we stub the auth section and auth plugin so
        # it can load the plugin specific options.
        loading.register_auth_conf_options(
            self.conf_fixture.conf, group=self.GROUP
        )

    def test_loading_v2(self):
        section = uuid.uuid4().hex
        auth_url = uuid.uuid4().hex
        username = uuid.uuid4().hex
        password = uuid.uuid4().hex
        trust_id = uuid.uuid4().hex
        tenant_id = uuid.uuid4().hex

        self.conf_fixture.config(auth_section=section, group=self.GROUP)
        loading.register_auth_conf_options(
            self.conf_fixture.conf, group=self.GROUP
        )

        opts = loading.get_auth_plugin_conf_options(v2.Password())
        self.conf_fixture.register_opts(opts, group=section)

        self.conf_fixture.config(
            auth_type=self.V2PASS,
            auth_url=auth_url,
            username=username,
            password=password,
            trust_id=trust_id,
            tenant_id=tenant_id,
            group=section,
        )

        a = loading.load_auth_from_conf_options(
            self.conf_fixture.conf, self.GROUP
        )

        self.assertEqual(auth_url, a.auth_url)
        self.assertEqual(username, a.username)
        self.assertEqual(password, a.password)
        self.assertEqual(trust_id, a.trust_id)
        self.assertEqual(tenant_id, a.tenant_id)

    def test_loading_v3(self):
        section = uuid.uuid4().hex
        auth_url = (uuid.uuid4().hex,)
        token = uuid.uuid4().hex
        trust_id = uuid.uuid4().hex
        project_id = uuid.uuid4().hex
        project_domain_name = uuid.uuid4().hex

        self.conf_fixture.config(auth_section=section, group=self.GROUP)
        loading.register_auth_conf_options(
            self.conf_fixture.conf, group=self.GROUP
        )

        opts = loading.get_auth_plugin_conf_options(v3.Token())
        self.conf_fixture.register_opts(opts, group=section)

        self.conf_fixture.config(
            auth_type=self.V3TOKEN,
            auth_url=auth_url,
            token=token,
            trust_id=trust_id,
            project_id=project_id,
            project_domain_name=project_domain_name,
            group=section,
        )

        a = loading.load_auth_from_conf_options(
            self.conf_fixture.conf, self.GROUP
        )

        self.assertEqual(token, a.auth_methods[0].token)
        self.assertEqual(trust_id, a.trust_id)
        self.assertEqual(project_id, a.project_id)
        self.assertEqual(project_domain_name, a.project_domain_name)

    def test_loading_invalid_plugin(self):
        auth_type = uuid.uuid4().hex
        self.conf_fixture.config(auth_type=auth_type, group=self.GROUP)

        e = self.assertRaises(
            exceptions.NoMatchingPlugin,
            loading.load_auth_from_conf_options,
            self.conf_fixture.conf,
            self.GROUP,
        )

        self.assertEqual(auth_type, e.name)

    def test_loading_with_no_data(self):
        lo = loading.load_auth_from_conf_options(
            self.conf_fixture.conf, self.GROUP
        )
        self.assertIsNone(lo)

    @mock.patch('stevedore.DriverManager')
    def test_other_params(self, m):
        m.return_value = utils.MockManager(utils.MockLoader())
        driver_name = uuid.uuid4().hex

        opts = loading.get_auth_plugin_conf_options(utils.MockLoader())
        self.conf_fixture.register_opts(opts, group=self.GROUP)
        self.conf_fixture.config(
            auth_type=driver_name, group=self.GROUP, **self.TEST_VALS
        )

        a = loading.load_auth_from_conf_options(
            self.conf_fixture.conf, self.GROUP
        )
        self.assertTestVals(a)

        m.assert_called_once_with(
            namespace=loading.PLUGIN_NAMESPACE,
            name=driver_name,
            invoke_on_load=True,
        )

    @utils.mock_plugin()
    def test_same_section(self, m):
        opts = loading.get_auth_plugin_conf_options(utils.MockLoader())
        self.conf_fixture.register_opts(opts, group=self.GROUP)

        loading.register_auth_conf_options(
            self.conf_fixture.conf, group=self.GROUP
        )
        self.conf_fixture.config(
            auth_type=uuid.uuid4().hex, group=self.GROUP, **self.TEST_VALS
        )

        a = loading.load_auth_from_conf_options(
            self.conf_fixture.conf, self.GROUP
        )
        self.assertTestVals(a)

    @utils.mock_plugin()
    def test_diff_section(self, m):
        section = uuid.uuid4().hex

        self.conf_fixture.config(auth_section=section, group=self.GROUP)
        loading.register_auth_conf_options(
            self.conf_fixture.conf, group=self.GROUP
        )

        opts = loading.get_auth_plugin_conf_options(utils.MockLoader())
        self.conf_fixture.register_opts(opts, group=section)
        self.conf_fixture.config(
            group=section, auth_type=uuid.uuid4().hex, **self.TEST_VALS
        )

        a = loading.load_auth_from_conf_options(
            self.conf_fixture.conf, self.GROUP
        )
        self.assertTestVals(a)

    def test_plugins_are_all_opts(self):
        manager = stevedore.ExtensionManager(
            loading.PLUGIN_NAMESPACE, propagate_map_exceptions=True
        )

        def inner(driver):
            for p in driver.plugin().get_options():
                self.assertIsInstance(p, loading.Opt)

        manager.map(inner)

    def test_get_common(self):
        opts = loading.get_auth_common_conf_options()
        for opt in opts:
            self.assertIsInstance(opt, cfg.Opt)
        self.assertEqual(2, len(opts))

    def test_get_named(self):
        loaded_opts = loading.get_plugin_options('v2password')
        plugin_opts = v2.Password().get_options()

        loaded_names = {o.name for o in loaded_opts}
        plugin_names = {o.name for o in plugin_opts}

        self.assertEqual(plugin_names, loaded_names)

    def test_register_cfg(self):
        loading.register_auth_conf_options(
            self.conf_fixture.conf, group=self.GROUP
        )

    def test_common_conf_options(self):
        opts = loading.get_auth_common_conf_options()

        self.assertEqual(2, len(opts))
        auth_type = [o for o in opts if o.name == 'auth_type'][0]
        self.assertEqual(1, len(auth_type.deprecated_opts))
        self.assertIsInstance(auth_type.deprecated_opts[0], cfg.DeprecatedOpt)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
import datetime
import functools
import hashlib
import json
import logging
import os
import platform
import socket
import sys
import time
import types
import typing as ty
import urllib
import uuid

import requests

import keystoneauth1
from keystoneauth1 import _utils as utils
from keystoneauth1 import discover
from keystoneauth1 import exceptions

try:
    import netaddr
except ImportError:
    netaddr = None

try:
    import osprofiler.web as osprofiler_web
except ImportError:
    osprofiler_web = None

if ty.TYPE_CHECKING:
    from keystoneauth1 import plugin
    import requests.auth

DEFAULT_USER_AGENT = f'keystoneauth1/{keystoneauth1.__version__} {requests.utils.default_user_agent()} {platform.python_implementation()}/{platform.python_version()}'

# NOTE(jamielennox): Clients will likely want to print more than json. Please
# propose a patch if you have a content type you think is reasonable to print
# here and we'll add it to the list as required.
_LOG_CONTENT_TYPES = {'application/json', 'text/plain'}

_MAX_RETRY_INTERVAL = 60.0
_EXPONENTIAL_DELAY_START = 0.5
_RETRIABLE_STATUS_CODES = [503]

# NOTE(efried): This is defined in oslo_middleware.request_id.INBOUND_HEADER,
# but it didn't seem worth adding oslo_middleware to requirements just for that
_REQUEST_ID_HEADER = 'X-Openstack-Request-Id'


def _construct_session(
    session_obj: requests.Session | None = None,
) -> requests.Session:
    # NOTE(morganfainberg): if the logic in this function changes be sure to
    # update the betamax fixture's '_construct_session_with_betamax" function
    # as well.
    if not session_obj:
        session_obj = requests.Session()
        # Use TCPKeepAliveAdapter to fix bug 1323862
        for scheme in list(session_obj.adapters):
            session_obj.mount(scheme, TCPKeepAliveAdapter())
    return session_obj


def _mv_legacy_headers_for_service(mv_service_type: str) -> list[str]:
    """Workaround for services that predate standardization.

    TODO(sdague): eventually convert this to using os-service-types
    and put the logic there. However, right now this is so little
    logic, inlining it for release is a better call.
    """
    headers = []
    if mv_service_type == "compute":
        headers.append("X-OpenStack-Nova-API-Version")
    elif mv_service_type == "baremetal":
        headers.append("X-OpenStack-Ironic-API-Version")
    elif mv_service_type in ["sharev2", "shared-file-system"]:
        headers.append("X-OpenStack-Manila-API-Version")
    return headers


def _sanitize_headers(headers: dict[str | bytes, ty.Any]) -> dict[str, ty.Any]:
    """Ensure headers are strings and not bytes."""
    str_dict = {}
    for k, v in headers.items():
        # requests expects headers to be str type in python3, which means
        # if we get a bytes we need to decode it into a str
        k = k.decode('ASCII') if isinstance(k, bytes) else k
        if v is not None:
            v = v.decode('ASCII') if isinstance(v, bytes) else v
        str_dict[k] = v
    return str_dict


class NoOpSemaphore:
    """Empty context manager for use as a default semaphore."""

    def __enter__(self) -> None:
        """Enter the context manager and do nothing."""
        pass

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        """Exit the context manager and do nothing."""
        pass


class _JSONEncoder(json.JSONEncoder):
    def default(self, o: object) -> ty.Any:
        if isinstance(o, datetime.datetime):
            return o.isoformat()
        if isinstance(o, uuid.UUID):
            return str(o)
        if netaddr and isinstance(o, netaddr.IPAddress):
            return str(o)

        return super().default(o)


class _StringFormatter:
    """A String formatter that fetches values on demand."""

    def __init__(
        self, session: 'Session', auth: ty.Optional['plugin.BaseAuthPlugin']
    ):
        self.session = session
        self.auth = auth

    def __getitem__(self, item: str) -> ty.Any:
        if item == 'project_id':
            value = self.session.get_project_id(self.auth)
        elif item == 'user_id':
            value = self.session.get_user_id(self.auth)
        else:
            raise AttributeError(item)

        if not value:
            raise ValueError(
                "This type of authentication does not provide a "
                f"{item} that can be substituted"
            )

        return value


def _determine_calling_package() -> str:
    """Walk the call frames trying to identify what is using this module."""
    # Create a lookup table mapping file name to module name. The ``inspect``
    # module does this but is far less efficient. Same story with the
    # frame walking below.  One could use ``inspect.stack()`` but it
    # has far more overhead.

    # NOTE(jwysogla): According to the docs, we should always use copy(),
    # because sys.modules can change during iteration, which results
    # in a RuntimeError
    # https://docs.python.org/3/library/sys.html#sys.modules
    mod_lookup = {
        m.__file__: n
        for n, m in sys.modules.copy().items()
        if hasattr(m, '__file__')
    }

    # NOTE(shaleh): these are not useful because they hide the real
    # user of the code. debtcollector did not import keystoneauth but
    # it will show up in the call stack. Similarly we do not want to
    # report ourselves or keystone client as the user agent. The real
    # user is the code importing them.
    ignored = ('debtcollector', 'keystoneauth1', 'keystoneclient')

    i = 0
    while True:
        i += 1

        try:
            # NOTE(shaleh): this is safe in CPython but could break in
            # other implementations of Python. Yes, the `inspect`
            # module could be used instead. But it does a lot more
            # work so it has worse performance.
            f = sys._getframe(i)
            try:
                name = mod_lookup[f.f_code.co_filename]
                # finds the full name module.foo.bar but all we need
                # is the module name.
                name, _, _ = name.partition('.')
                if name not in ignored:
                    return name
            except KeyError:
                pass  # builtin or the like
        except ValueError:
            # hit the bottom of the frame stack
            break

    return ''


def _determine_user_agent() -> str:
    """Attempt to programmatically generate a user agent string.

    First, look at the name of the process. Return this unless it is in
    the `ignored` list.  Otherwise, look at the function call stack and
    try to find the name of the code that invoked this module.
    """
    # NOTE(shaleh): mod_wsgi is not any more useful than just
    # reporting "keystoneauth". Ignore it and perform the package name
    # heuristic.
    ignored = ('mod_wsgi',)

    try:
        name = sys.argv[0]
    except IndexError:
        # sys.argv is empty, usually the Python interpreter prevents this.
        return ''

    if not name:
        return ''

    name = os.path.basename(name)
    if name in ignored:
        name = _determine_calling_package()
    return name


class RequestTiming:
    """Contains timing information for an HTTP interaction."""

    #: HTTP method used for the call (GET, POST, etc)
    method: str | None

    #: URL against which the call was made
    url: str | None

    #: Elapsed time information
    elapsed: datetime.timedelta

    def __init__(
        self, method: str | None, url: str | None, elapsed: datetime.timedelta
    ):
        self.method = method
        self.url = url
        self.elapsed = elapsed


class _Retries:
    __slots__ = ('_fixed_delay', '_current')
    _fixed_delay: float | None
    _current: float

    def __init__(self, fixed_delay: float | None = None):
        self._fixed_delay = fixed_delay
        self.reset()

    def __next__(self) -> float:
        value = self._current
        if not self._fixed_delay:
            self._current = min(value * 2, _MAX_RETRY_INTERVAL)
        return value

    def reset(self) -> None:
        if self._fixed_delay:
            self._current = self._fixed_delay
        else:
            self._current = _EXPONENTIAL_DELAY_START


class Session:
    """Maintains client communication state and common functionality.

    As much as possible the parameters to this class reflect and are passed
    directly to the :mod:`requests` library.

    :param auth: An authentication plugin to authenticate the session with.
                 (optional, defaults to None)
    :type auth: keystoneauth1.plugin.BaseAuthPlugin
    :param requests.Session session: A requests session object that can be used
                                     for issuing requests. (optional)
    :param str original_ip: The original IP of the requesting user which will
                            be sent to identity service in a 'Forwarded'
                            header. (optional)
    :param verify: The verification arguments to pass to requests. These are of
                   the same form as requests expects, so True or False to
                   verify (or not) against system certificates or a path to a
                   bundle or CA certs to check against or None for requests to
                   attempt to locate and use certificates. (optional, defaults
                   to True)
    :param cert: A client certificate to pass to requests. These are of the
                 same form as requests expects. Either a single filename
                 containing both the certificate and key or a tuple containing
                 the path to the certificate then a path to the key. (optional)
    :param float timeout: A timeout to pass to requests. This should be a
                          numerical value indicating some amount (or fraction)
                          of seconds or 0 for no timeout. (optional, defaults
                          to 0)
    :param str user_agent: A User-Agent header string to use for the request.
                           If not provided, a default of
                           :attr:`~keystoneauth1.session.DEFAULT_USER_AGENT` is
                           used, which contains the keystoneauth1 version as
                           well as those of the requests library and which
                           Python is being used. When a non-None value is
                           passed, it will be prepended to the default.
    :param int/bool redirect: Controls the maximum number of redirections that
                              can be followed by a request. Either an integer
                              for a specific count or True/False for
                              forever/never. (optional, default to 30)
    :param dict additional_headers: Additional headers that should be attached
                                    to every request passing through the
                                    session. Headers of the same name specified
                                    per request will take priority.
    :param str app_name: The name of the application that is creating the
                         session. This will be used to create the user_agent.
    :param str app_version: The version of the application creating the
                            session. This will be used to create the
                            user_agent.
    :param list additional_user_agent: A list of tuple of name, version that
                                       will be added to the user agent. This
                                       can be used by libraries that are part
                                       of the communication process.
    :param dict discovery_cache: A dict to be used for caching of discovery
                                 information. This is normally managed
                                 transparently, but if the user wants to
                                 share a single cache across multiple sessions
                                 that do not share an auth plugin, it can
                                 be provided here. (optional, defaults to
                                 None which means automatically manage)
    :param bool split_loggers: Split the logging of requests across multiple
                               loggers instead of just one. Defaults to False.
    :param bool collect_timing: Whether or not to collect per-method timing
                                information for each API call. (optional,
                                defaults to False)
    :param rate_semaphore: Semaphore to be used to control concurrency
                           and rate limiting of requests. (optional,
                           defaults to no concurrency or rate control)
    :param int connect_retries: the maximum number of retries that should
                                be attempted for connection errors.
                                (optional, defaults to 0 - never retry).
    """

    user_agent = None

    _REDIRECT_STATUSES = (301, 302, 303, 305, 307, 308)

    _DEFAULT_REDIRECT_LIMIT = 30

    def __init__(
        self,
        auth: ty.Optional['plugin.BaseAuthPlugin'] = None,
        session: requests.Session | None = None,
        original_ip: str | None = None,
        verify: bool | str | None = True,
        cert: str | tuple[str, str] | None = None,
        timeout: int | None = None,
        user_agent: str | None = None,
        redirect: int | bool = _DEFAULT_REDIRECT_LIMIT,
        additional_headers: collections.abc.MutableMapping[str, str]
        | None = None,
        app_name: str | None = None,
        app_version: str | None = None,
        additional_user_agent: list[tuple[str, str]] | None = None,
        discovery_cache: dict[str, ty.Any] | None = None,
        split_loggers: bool | None = None,
        collect_timing: bool = False,
        rate_semaphore: ty.ContextManager[None] | None = None,
        connect_retries: int = 0,
    ):
        self.auth = auth
        self.session = _construct_session(session)
        # NOTE(mwhahaha): keep a reference to the session object so we can
        # clean it up when this object goes away. We don't want to close the
        # session if it was passed into us as it may be reused externally.
        # See LP#1838704
        self._session: requests.Session | None = None
        if not session:
            self._session = self.session
        self.original_ip = original_ip
        self.verify = verify
        self.cert = cert
        self.timeout = None
        self.redirect = redirect
        self.additional_headers = additional_headers or {}
        self.app_name = app_name
        self.app_version = app_version
        self.additional_user_agent = additional_user_agent or []
        self._determined_user_agent: str | None = None
        if discovery_cache is None:
            discovery_cache = {}
        self._discovery_cache = discovery_cache
        # NOTE(mordred) split_loggers kwarg default is None rather than False
        # so we can distinguish between the value being set or not.
        self._split_loggers = split_loggers
        self._collect_timing = collect_timing
        self._connect_retries = connect_retries
        self._api_times: list[RequestTiming] = []
        self._rate_semaphore = rate_semaphore or NoOpSemaphore()

        if timeout is not None:
            self.timeout = float(timeout)

        if user_agent is not None:
            self.user_agent = f"{user_agent} {DEFAULT_USER_AGENT}"

        self._json = _JSONEncoder()

    def __del__(self) -> None:
        """Clean up resources on delete."""
        if self._session:
            # If we created a requests.Session, try to close it out correctly
            try:
                self._session.close()
            except Exception:
                pass
            finally:
                self._session = None

    # TODO(stephenfin): The types from typeshed are very generic. This would be
    # more accurate:
    #
    #     collections.OrderedDict[str, requests.adapters.BaseAdapter]
    @property
    def adapters(self) -> ty.MutableMapping[ty.Any, ty.Any]:
        return self.session.adapters

    @adapters.setter
    def adapters(
        self,
        value: collections.OrderedDict[str, requests.adapters.BaseAdapter],
    ) -> None:
        self.session.adapters = value

    def mount(
        self, scheme: str, adapter: requests.adapters.BaseAdapter
    ) -> None:
        self.session.mount(scheme, adapter)

    def _remove_service_catalog(self, body: str) -> str:
        try:
            data = json.loads(body)

            # V3 token
            if 'token' in data and 'catalog' in data['token']:
                data['token']['catalog'] = '<removed>'
                return self._json.encode(data)

            # V2 token
            if 'serviceCatalog' in data['access']:
                data['access']['serviceCatalog'] = '<removed>'
                return self._json.encode(data)

        except Exception:
            # Don't fail trying to clean up the request body.
            pass
        return body

    @staticmethod
    def _process_header(header: tuple[str, str]) -> tuple[str, str]:
        """Redact the secure headers to be logged."""
        secure_headers = (
            'authorization',
            'x-auth-token',
            'x-subject-token',
            'x-service-token',
        )
        if header[0].lower() in secure_headers:
            token_hasher = hashlib.sha256()
            token_hasher.update(header[1].encode('utf-8'))
            token_hash = token_hasher.hexdigest()
            return (header[0], f'{{SHA256}}{token_hash}')
        return header

    def _get_split_loggers(self, split_loggers: bool | None) -> bool | None:
        """Get a boolean value from the various argument sources.

        We default split_loggers to None in the kwargs of the Session
        constructor so we can track set vs. not set. We also accept
        split_loggers as a parameter in a few other places. In each place
        we want the parameter, if given by the user, to win.
        """
        # None is the default value in each method's kwarg. None means "unset".
        if split_loggers is None:
            # If no value was given, try the value set on the instance.
            split_loggers = self._split_loggers
        if split_loggers is None:
            # If neither a value was given on the method, nor a value was given
            # on the Session constructor, then the value is False.
            split_loggers = False
        return split_loggers

    def _http_log_request(
        self,
        url: str,
        method: str | None = None,
        data: str | bytes | None = None,
        json: object = None,
        headers: collections.abc.MutableMapping[str, str] | None = None,
        query_params: dict[str, ty.Any] | None = None,
        logger: logging.Logger | None = None,
        split_loggers: bool | None = None,
    ) -> None:
        string_parts = []

        if self._get_split_loggers(split_loggers):
            _logger = utils.get_logger(__name__ + '.request')
        else:
            # Only a single logger was passed in, use string prefixing.
            string_parts.append('REQ:')
            if not logger:
                _logger = utils.get_logger(__name__)
            else:
                _logger = logger

        if not _logger.isEnabledFor(logging.DEBUG):
            # NOTE(morganfainberg): This whole debug section is expensive,
            # there is no need to do the work if we're not going to emit a
            # debug log.
            return

        string_parts.append('curl -g -i')

        # NOTE(jamielennox): None means let requests do its default validation
        # so we need to actually check that this is False.
        if self.verify is False:
            string_parts.append('--insecure')
        elif isinstance(self.verify, str):
            string_parts.append(f'--cacert "{self.verify}"')

        if method:
            string_parts.extend(['-X', method])

        if query_params:
            # Don't check against `is not None` as this can be
            # an empty dictionary, which we shouldn't bother logging.
            url = url + '?' + urllib.parse.urlencode(query_params)
            # URLs with query strings need to be wrapped in quotes in order
            # for the CURL command to run properly.
            string_parts.append(f'"{url}"')
        else:
            string_parts.append(url)

        if headers:
            # Sort headers so that testing can work consistently.
            for header in sorted(headers.items()):
                string_parts.append(
                    '-H "{}: {}"'.format(*self._process_header(header))
                )
        if json:
            data = self._json.encode(json)
        if data:
            if isinstance(data, bytes):
                try:
                    data = data.decode("ascii")
                except UnicodeDecodeError:
                    data = "<binary_data>"
            string_parts.append(f"-d '{data}'")

        _logger.debug(' '.join(string_parts))

    def _http_log_response(
        self,
        response: requests.Response | None = None,
        json: object = None,
        status_code: int | None = None,
        headers: collections.abc.MutableMapping[str, str] | None = None,
        text: str | None = None,
        *,
        logger: logging.Logger,
        split_loggers: bool | None = None,
    ) -> None:
        string_parts = []
        body_parts = []
        if self._get_split_loggers(split_loggers):
            logger = utils.get_logger(__name__ + '.response')
            body_logger = utils.get_logger(__name__ + '.body')
        else:
            # Only a single logger was passed in, use string prefixing.
            string_parts.append('RESP:')
            body_parts.append('RESP BODY:')
            body_logger = logger

        if not logger.isEnabledFor(logging.DEBUG):
            return

        if response is not None:
            if not status_code:
                status_code = response.status_code
            if not headers:
                headers = response.headers

        if status_code:
            string_parts.append(f'[{status_code}]')
        if headers:
            # Sort headers so that testing can work consistently.
            for header in sorted(headers.items()):
                string_parts.append(
                    '{}: {}'.format(*self._process_header(header))
                )
        logger.debug(' '.join(string_parts))

        if not body_logger.isEnabledFor(logging.DEBUG):
            return

        if response is not None:
            if not text:
                # NOTE(samueldmq): If the response does not provide enough info
                # about the content type to decide whether it is useful and
                # safe to log it or not, just do not log the body. Trying to
                # read the response body anyways may result on reading a long
                # stream of bytes and getting an unexpected MemoryError. See
                # bug 1616105 for further details.
                content_type = response.headers.get('content-type', None)

                # NOTE(lamt): Per [1], the Content-Type header can be of the
                # form Content-Type := type "/" subtype *[";" parameter]
                # [1] https://www.w3.org/Protocols/rfc1341/4_Content-Type.html
                for log_type in _LOG_CONTENT_TYPES:
                    if content_type is not None and content_type.startswith(
                        log_type
                    ):
                        text = self._remove_service_catalog(response.text)
                        break
                else:
                    text = (
                        'Omitted, Content-Type is set to %s. Only '
                        '%s responses have their bodies logged.'
                    )
                    text = text % (content_type, ', '.join(_LOG_CONTENT_TYPES))
        if json:
            text = self._json.encode(json)

        if text:
            body_parts.append(text)
            body_logger.debug(' '.join(body_parts))

    @staticmethod
    def _set_microversion_headers(
        headers: collections.abc.MutableMapping[str, str],
        microversion: str,
        service_type: str | None,
        endpoint_filter: dict[str, ty.Any] | None,
    ) -> None:
        # We're converting it to normalized version number for two reasons.
        # First, to validate it's a real version number. Second, so that in
        # the future we can pre-validate that it is within the range of
        # available microversions before we send the request.
        # TODO(mordred) Validate when we get the response back that
        # the server executed in the microversion we expected.
        # TODO(mordred) Validate that the requested microversion works
        # with the microversion range we found in discovery.
        parsed_microversion = discover.normalize_version_number(microversion)
        # Can't specify a M.latest microversion
        if (
            parsed_microversion[0] != discover.LATEST
            and discover.LATEST in parsed_microversion[1:]
        ):
            raise TypeError(
                "Specifying a '{major}.latest' microversion is not allowed."
            )

        microversion = discover.version_to_string(parsed_microversion)
        if not service_type:
            if endpoint_filter and 'service_type' in endpoint_filter:
                service_type = endpoint_filter['service_type']
            else:
                raise TypeError(
                    f"microversion {microversion} was requested but no"
                    " service_type information is available. Either provide a"
                    " service_type in endpoint_filter or pass"
                    " microversion_service_type as an argument."
                )

        service_type = ty.cast(str, service_type)  # narrow the type

        # TODO(mordred) cinder uses volume in its microversion header. This
        # logic should be handled in the future by os-service-types but for
        # now hard-code for cinder.
        if (
            service_type.startswith('volume')
            or service_type == 'block-storage'
        ):
            service_type = 'volume'
        elif service_type.startswith('share'):
            # NOTE(gouthamr) manila doesn't honor the "OpenStack-API-Version"
            # header yet, but sending it does no harm - when the service
            # honors this header, it'll use the standardized name in the
            # service-types-authority and not the legacy name in the cloud's
            # service catalog
            service_type = 'shared-file-system'

        headers.setdefault(
            'OpenStack-API-Version', f'{service_type} {microversion}'
        )
        header_names = _mv_legacy_headers_for_service(service_type)
        for h in header_names:
            headers.setdefault(h, microversion)

    def request(
        self,
        url: str,
        method: str,
        json: object | None = None,
        original_ip: str | None = None,
        user_agent: str | None = None,
        redirect: int | bool | None = None,
        authenticated: bool | None = None,
        endpoint_filter: dict[str, ty.Any] | None = None,
        auth: ty.Optional['plugin.BaseAuthPlugin'] = None,
        requests_auth: ty.Optional['requests.auth.AuthBase'] = None,
        raise_exc: bool = True,
        allow_reauth: bool = True,
        log: bool = True,
        endpoint_override: str | None = None,
        connect_retries: int | None = None,
        logger: logging.Logger | None = None,
        allow: dict[str, ty.Any] | None = None,
        client_name: str | None = None,
        client_version: str | None = None,
        microversion: str | None = None,
        microversion_service_type: str | None = None,
        status_code_retries: int = 0,
        retriable_status_codes: list[int] | None = None,
        rate_semaphore: ty.ContextManager[None] | None = None,
        global_request_id: str | None = None,
        connect_retry_delay: float | None = None,
        status_code_retry_delay: float | None = None,
        **kwargs: ty.Any,
    ) -> requests.Response:
        """Send an HTTP request with the specified characteristics.

        Wrapper around `requests.Session.request` to handle tasks such as
        setting headers, JSON encoding/decoding, and error handling.

        Arguments that are not handled are passed through to the requests
        library.

        :param str url: Path or fully qualified URL of HTTP request. If only a
                        path is provided then endpoint_filter must also be
                        provided such that the base URL can be determined. If a
                        fully qualified URL is provided then endpoint_filter
                        will be ignored.
        :param str method: The http method to use. (e.g. 'GET', 'POST')
        :param json: Some data to be represented as JSON. (optional)
        :param str original_ip: Mark this request as forwarded for this ip.
                                (optional)
        :param dict headers: Headers to be included in the request. (optional)
        :param str user_agent: A user_agent to use for the request. If present
                               will override one present in headers. (optional)
        :param int/bool redirect: the maximum number of redirections that
                                  can be followed by a request. Either an
                                  integer for a specific count or True/False
                                  for forever/never. (optional)
        :param int connect_retries: the maximum number of retries that should
                                    be attempted for connection errors.
                                    (optional, defaults to None - never retry).
        :param bool authenticated: True if a token should be attached to this
                                   request, False if not or None for attach if
                                   an auth_plugin is available.
                                   (optional, defaults to None)
        :param dict endpoint_filter: Data to be provided to an auth plugin with
                                     which it should be able to determine an
                                     endpoint to use for this request. If not
                                     provided then URL is expected to be a
                                     fully qualified URL. (optional)
        :param str endpoint_override: The URL to use instead of looking up the
                                      endpoint in the auth plugin. This will be
                                      ignored if a fully qualified URL is
                                      provided but take priority over an
                                      endpoint_filter. This string may contain
                                      the values ``%(project_id)s`` and
                                      ``%(user_id)s`` to have those values
                                      replaced by the project_id/user_id of the
                                      current authentication. (optional)
        :param auth: The auth plugin to use when authenticating this request.
                     This will override the plugin that is attached to the
                     session (if any). (optional)
        :type auth: keystoneauth1.plugin.BaseAuthPlugin
        :param requests_auth: A requests library auth plugin that cannot be
                              passed via kwarg because the `auth` kwarg
                              collides with our own auth plugins. (optional)
        :type requests_auth: :py:class:`requests.auth.AuthBase`
        :param bool raise_exc: If True then raise an appropriate exception for
                               failed HTTP requests. If False then return the
                               request object. (optional, default True)
        :param bool allow_reauth: Allow fetching a new token and retrying the
                                  request on receiving a 401 Unauthorized
                                  response. (optional, default True)
        :param bool log: If True then log the request and response data to the
                         debug log. (optional, default True)
        :param logger: The logger object to use to log request and responses.
                       If not provided the keystoneauth1.session default
                       logger will be used.
        :type logger: logging.Logger
        :param dict allow: Extra filters to pass when discovering API
                           versions. (optional)
        :param microversion: Microversion to send for this request.
                       microversion can be given as a string or a tuple.
                       (optional)
        :param str microversion_service_type: The service_type to be sent in
                       the microversion header, if a microversion is given.
                       Defaults to the value of service_type from
                       endpoint_filter if one exists. If endpoint_filter is not
                       provided or does not have a service_type, microversion
                       is given and microversion_service_type is not provided,
                       an exception will be raised.
        :param int status_code_retries: the maximum number of retries that
                                        should be attempted for retriable
                                        HTTP status codes (optional, defaults
                                        to 0 - never retry).
        :param list retriable_status_codes: list of HTTP status codes that
                                            should be retried (optional,
                                            defaults to HTTP 503, has no effect
                                            when status_code_retries is 0).
        :param rate_semaphore: Semaphore to be used to control concurrency
                               and rate limiting of requests. (optional,
                               defaults to no concurrency or rate control)
        :param global_request_id: Value for the X-Openstack-Request-Id header.
        :param float connect_retry_delay: Delay (in seconds) between two
                                          connect retries (if enabled).
                                          By default exponential retry starting
                                          with 0.5 seconds up to a maximum of
                                          60 seconds is used.
        :param float status_code_retry_delay: Delay (in seconds) between two
                                              status code retries (if enabled).
                                              By default exponential retry
                                              starting with 0.5 seconds up to
                                              a maximum of 60 seconds is used.
        :param kwargs: any other parameter that can be passed to
                       :meth:`requests.Session.request` (such as `headers`).
                       Except:

                       - `data` will be overwritten by the data in the `json`
                         param.
                       - `allow_redirects` is ignored as redirects are handled
                         by the session.

        :raises keystoneauth1.exceptions.base.ClientException: For connection
            failure, or to indicate an error response code.

        :returns: The response to the request.
        """
        # If a logger is passed in, use it and do not log requests, responses
        # and bodies separately.
        if logger:
            split_loggers = False
        else:
            split_loggers = None
        logger = logger or utils.get_logger(__name__)
        # NOTE(gmann): Convert r initlize the headers to
        # CaseInsensitiveDict to make sure headers are
        # case insensitive.
        if kwargs.get('headers'):
            kwargs['headers'] = requests.structures.CaseInsensitiveDict(
                kwargs['headers']
            )
        else:
            kwargs['headers'] = requests.structures.CaseInsensitiveDict()
        if connect_retries is None:
            connect_retries = self._connect_retries
        # HTTP 503 - Service Unavailable
        retriable_status_codes = (
            retriable_status_codes or _RETRIABLE_STATUS_CODES
        )
        rate_semaphore = rate_semaphore or self._rate_semaphore

        headers = kwargs.setdefault('headers', {})
        if microversion:
            self._set_microversion_headers(
                headers,
                microversion,
                microversion_service_type,
                endpoint_filter,
            )

        if authenticated is None:
            authenticated = bool(auth or self.auth)

        if authenticated:
            auth_headers = self.get_auth_headers(auth)

            if auth_headers is None:
                msg = 'No valid authentication is available'
                raise exceptions.AuthorizationFailure(msg)

            headers.update(auth_headers)

        if osprofiler_web:
            headers.update(osprofiler_web.get_trace_id_headers())

        # if we are passed a fully qualified URL and an endpoint_filter we
        # should ignore the filter. This will make it easier for clients who
        # want to overrule the default endpoint_filter data added to all client
        # requests. We check fully qualified here by the presence of a host.
        if not urllib.parse.urlparse(url).netloc:
            base_url = None

            if endpoint_override:
                base_url = endpoint_override % _StringFormatter(self, auth)
            elif endpoint_filter:
                base_url = self.get_endpoint(
                    auth, allow=allow, **endpoint_filter
                )

            if not base_url:
                raise exceptions.EndpointNotFound()

            url = '{}/{}'.format(base_url.rstrip('/'), url.lstrip('/'))

        if self.cert:
            kwargs.setdefault('cert', self.cert)

        if self.timeout is not None:
            kwargs.setdefault('timeout', self.timeout)

        if user_agent:
            headers['User-Agent'] = user_agent
        elif self.user_agent:
            user_agent = headers.setdefault('User-Agent', self.user_agent)
        else:
            # Per RFC 7231 Section 5.5.3, identifiers in a user-agent should be
            # ordered by decreasing significance.  If a user sets their product
            # that value will be used. Otherwise we attempt to derive a useful
            # product value. The value will be prepended it to the KSA version,
            # requests version, and then the Python version.

            agent = []

            if self.app_name and self.app_version:
                agent.append(f'{self.app_name}/{self.app_version}')
            elif self.app_name:
                agent.append(self.app_name)

            if client_name and client_version:
                agent.append(f'{client_name}/{client_version}')
            elif client_name:
                agent.append(client_name)

            for additional in self.additional_user_agent:
                agent.append('{}/{}'.format(*additional))

            if not agent:
                # NOTE(jamielennox): determine_user_agent will return an empty
                # string on failure so checking for None will ensure it is only
                # called once even on failure.
                if self._determined_user_agent is None:
                    self._determined_user_agent = _determine_user_agent()

                if self._determined_user_agent:
                    agent.append(self._determined_user_agent)

            agent.append(DEFAULT_USER_AGENT)
            user_agent = headers.setdefault('User-Agent', ' '.join(agent))

        if self.original_ip:
            headers.setdefault(
                'Forwarded', f'for={self.original_ip};by={user_agent}'
            )

        if json is not None:
            headers.setdefault('Content-Type', 'application/json')
            kwargs['data'] = self._json.encode(json)

        if global_request_id is not None:
            # NOTE(efried): This does *not* setdefault. If a global_request_id
            # kwarg was explicitly specified, it should override any value
            # previously configured (e.g. in Adapter.global_request_id).
            headers[_REQUEST_ID_HEADER] = global_request_id

        for k, v in self.additional_headers.items():
            headers.setdefault(k, v)

        # Bug #1766235: some headers may be bytes
        headers = _sanitize_headers(headers)
        kwargs['headers'] = headers

        kwargs.setdefault('verify', self.verify)

        if requests_auth:
            kwargs['auth'] = requests_auth

        # Query parameters that are included in the url string will
        # be logged properly, but those sent in the `params` parameter
        # (which the requests library handles) need to be explicitly
        # picked out so they can be included in the URL that gets loggged.
        query_params = kwargs.get('params', {})

        if log:
            self._http_log_request(
                url,
                method=method,
                data=kwargs.get('data'),
                headers=headers,
                query_params=query_params,
                logger=logger,
                split_loggers=split_loggers,
            )

        # Force disable requests redirect handling. We will manage this below.
        kwargs['allow_redirects'] = False

        if redirect is None:
            redirect = self.redirect

        connect_retry_delays = _Retries(connect_retry_delay)
        status_code_retry_delays = _Retries(status_code_retry_delay)

        send = functools.partial(
            self._send_request,
            url,
            method,
            redirect,
            log,
            logger,
            split_loggers,
            connect_retries,
            status_code_retries,
            retriable_status_codes,
            rate_semaphore,
            connect_retry_delays,
            status_code_retry_delays,
        )

        try:
            connection_params = self.get_auth_connection_params(auth=auth)
        except exceptions.MissingAuthPlugin:
            # NOTE(jamielennox): If we've gotten this far without an auth
            # plugin then we should be happy with allowing no additional
            # connection params. This will be the typical case for plugins
            # anyway.
            pass
        else:
            if connection_params:
                kwargs.update(connection_params)

        resp = send(**kwargs)

        # log callee and caller request-id for each api call
        if log:
            # service_name should be fetched from endpoint_filter if it is not
            # present then use service_type as service_name.
            service_name = None
            if endpoint_filter:
                service_name = endpoint_filter.get('service_name')
                if not service_name:
                    service_name = endpoint_filter.get('service_type')

            # Nova uses 'x-compute-request-id' and other services like
            # Glance, Cinder etc are using 'x-openstack-request-id' to store
            # request-id in the header
            request_id = resp.headers.get(
                'x-openstack-request-id'
            ) or resp.headers.get('x-compute-request-id')
            if request_id:
                if self._get_split_loggers(split_loggers):
                    id_logger = utils.get_logger(__name__ + '.request-id')
                else:
                    id_logger = logger
                if service_name:
                    id_logger.debug(
                        '%(method)s call to %(service_name)s for '
                        '%(url)s used request id '
                        '%(response_request_id)s',
                        {
                            'method': resp.request.method,
                            'service_name': service_name,
                            'url': resp.url,
                            'response_request_id': request_id,
                        },
                    )
                else:
                    id_logger.debug(
                        '%(method)s call to '
                        '%(url)s used request id '
                        '%(response_request_id)s',
                        {
                            'method': resp.request.method,
                            'url': resp.url,
                            'response_request_id': request_id,
                        },
                    )

        # handle getting a 401 Unauthorized response by invalidating the plugin
        # and then retrying the request. This is only tried once.
        if resp.status_code == 401 and authenticated and allow_reauth:
            if self.invalidate(auth):
                auth_headers = self.get_auth_headers(auth)

                if auth_headers is not None:
                    headers.update(auth_headers)
                    resp = send(**kwargs)

        if raise_exc and resp.status_code >= 400:
            logger.debug(
                'Request returned failure status: %s', resp.status_code
            )
            raise exceptions.from_response(resp, method, url)

        if self._collect_timing:
            for h in resp.history:
                self._api_times.append(
                    RequestTiming(
                        method=h.request.method,
                        url=h.request.url,
                        elapsed=h.elapsed,
                    )
                )
            self._api_times.append(
                RequestTiming(
                    method=resp.request.method,
                    url=resp.request.url,
                    elapsed=resp.elapsed,
                )
            )

        return resp

    def _send_request(
        self,
        url: str,
        method: str,
        redirect: int | bool,
        log: bool,
        logger: logging.Logger,
        split_loggers: bool | None,
        connect_retries: int,
        status_code_retries: int,
        retriable_status_codes: list[int],
        rate_semaphore: ty.ContextManager[None],
        connect_retry_delays: _Retries,
        status_code_retry_delays: _Retries,
        **kwargs: ty.Any,
    ) -> requests.Response:
        # NOTE(jamielennox): We handle redirection manually because the
        # requests lib follows some browser patterns where it will redirect
        # POSTs as GETs for certain statuses which is not want we want for an
        # API. See: https://en.wikipedia.org/wiki/Post/Redirect/Get

        # NOTE(jamielennox): The interaction between retries and redirects are
        # handled naively. We will attempt only a maximum number of retries and
        # redirects rather than per request limits. Otherwise the extreme case
        # could be redirects * retries requests. This will be sufficient in
        # most cases and can be fixed properly if there's ever a need.

        try:
            try:
                with rate_semaphore:
                    resp = self.session.request(method, url, **kwargs)
            except requests.exceptions.SSLError as e:
                msg = f'SSL exception connecting to {url}: {e}'
                raise exceptions.SSLError(msg)
            except requests.exceptions.Timeout:
                msg = f'Request to {url} timed out'
                raise exceptions.ConnectTimeout(msg)
            except requests.exceptions.ConnectionError as e:
                # NOTE(sdague): urllib3/requests connection error is a
                # translation of SocketError. However, SocketError
                # happens for many different reasons, and that low
                # level message is often really important in figuring
                # out the difference between network misconfigurations
                # and firewall blocking.
                msg = f'Unable to establish connection to {url}: {e}'
                raise exceptions.ConnectFailure(msg)
            except requests.exceptions.RequestException as e:
                msg = f'Unexpected exception for {url}: {e}'
                raise exceptions.UnknownConnectionError(msg, e)

        except exceptions.RetriableConnectionFailure as e:
            if connect_retries <= 0:
                raise

            delay = next(connect_retry_delays)
            logger.warning(
                'Failure: %(e)s. Retrying in %(delay).1fs. '
                '%(retries)s retries left',
                {'e': e, 'delay': delay, 'retries': connect_retries},
            )
            time.sleep(delay)

            return self._send_request(
                url,
                method,
                redirect,
                log,
                logger,
                split_loggers,
                status_code_retries=status_code_retries,
                retriable_status_codes=retriable_status_codes,
                rate_semaphore=rate_semaphore,
                connect_retries=connect_retries - 1,
                connect_retry_delays=connect_retry_delays,
                status_code_retry_delays=status_code_retry_delays,
                **kwargs,
            )

        if log:
            self._http_log_response(
                response=resp, logger=logger, split_loggers=split_loggers
            )

        if resp.status_code in self._REDIRECT_STATUSES:
            # be careful here in python True == 1 and False == 0
            if isinstance(redirect, bool):
                redirect_allowed = redirect
            else:
                redirect -= 1
                redirect_allowed = redirect >= 0

            if not redirect_allowed:
                return resp

            try:
                location = resp.headers['location']
            except KeyError:
                logger.warning(
                    "Failed to redirect request to %s as new "
                    "location was not provided.",
                    resp.url,
                )
            else:
                # NOTE(TheJulia): Location redirects generally should have
                # URI's to the destination.
                # https://datatracker.ietf.org/doc/html/rfc7231#section-7.1.2
                if 'params' in kwargs:
                    kwargs['params'] = {}

                if 'x-openstack-request-id' in resp.headers:
                    kwargs['headers'].setdefault(
                        'x-openstack-request-id',
                        resp.headers['x-openstack-request-id'],
                    )

                # NOTE(jamielennox): We don't keep increasing delays.
                # This request actually worked so we can reset the delay count.
                connect_retry_delays.reset()
                status_code_retry_delays.reset()
                new_resp = self._send_request(
                    location,
                    method,
                    redirect,
                    log,
                    logger,
                    split_loggers,
                    rate_semaphore=rate_semaphore,
                    connect_retries=connect_retries,
                    status_code_retries=status_code_retries,
                    retriable_status_codes=retriable_status_codes,
                    connect_retry_delays=connect_retry_delays,
                    status_code_retry_delays=status_code_retry_delays,
                    **kwargs,
                )

                if not isinstance(new_resp.history, list):
                    new_resp.history = list(new_resp.history)
                new_resp.history.insert(0, resp)
                resp = new_resp
        elif (
            resp.status_code in retriable_status_codes
            and status_code_retries > 0
        ):
            delay = next(status_code_retry_delays)
            logger.warning(
                'Retriable status code %(code)s. Retrying in '
                '%(delay).1fs. %(retries)s retries left',
                {
                    'code': resp.status_code,
                    'delay': delay,
                    'retries': status_code_retries,
                },
            )
            time.sleep(delay)

            # NOTE(jamielennox): We don't keep increasing connection delays.
            # This request actually worked so we can reset the delay count.
            connect_retry_delays.reset()
            return self._send_request(
                url,
                method,
                redirect,
                log,
                logger,
                split_loggers,
                connect_retries=connect_retries,
                status_code_retries=status_code_retries - 1,
                retriable_status_codes=retriable_status_codes,
                rate_semaphore=rate_semaphore,
                connect_retry_delays=connect_retry_delays,
                status_code_retry_delays=status_code_retry_delays,
                **kwargs,
            )

        return resp

    def get(self, url: str, **kwargs: ty.Any) -> requests.Response:
        """Perform a GET request.

        This calls :py:meth:`.request()` with ``method`` set to ``GET``.
        """
        return self.request(url, 'GET', **kwargs)

    def head(self, url: str, **kwargs: ty.Any) -> requests.Response:
        """Perform a HEAD request.

        This calls :py:meth:`.request()` with ``method`` set to ``HEAD``.
        """
        return self.request(url, 'HEAD', **kwargs)

    def post(self, url: str, **kwargs: ty.Any) -> requests.Response:
        """Perform a POST request.

        This calls :py:meth:`.request()` with ``method`` set to ``POST``.
        """
        return self.request(url, 'POST', **kwargs)

    def put(self, url: str, **kwargs: ty.Any) -> requests.Response:
        """Perform a PUT request.

        This calls :py:meth:`.request()` with ``method`` set to ``PUT``.
        """
        return self.request(url, 'PUT', **kwargs)

    def patch(self, url: str, **kwargs: ty.Any) -> requests.Response:
        """Perform a PATCH request.

        This calls :py:meth:`.request()` with ``method`` set to ``PATCH``.
        """
        return self.request(url, 'PATCH', **kwargs)

    def delete(self, url: str, **kwargs: ty.Any) -> requests.Response:
        """Perform a DELETE request.

        This calls :py:meth:`.request()` with ``method`` set to ``DELETE``.
        """
        return self.request(url, 'DELETE', **kwargs)

    def _auth_required(
        self, auth: ty.Optional['plugin.BaseAuthPlugin'], msg: str
    ) -> 'plugin.BaseAuthPlugin':
        if not auth:
            auth = self.auth

        if not auth:
            msg_fmt = 'An auth plugin is required to %s'
            raise exceptions.MissingAuthPlugin(msg_fmt % msg)

        return auth

    def get_auth_headers(
        self, auth: ty.Optional['plugin.BaseAuthPlugin'] = None
    ) -> dict[str, str] | None:
        """Return auth headers as provided by the auth plugin.

        :param auth: The auth plugin to use for token. Overrides the plugin
                     on the session. (optional)
        :type auth: keystoneauth1.plugin.BaseAuthPlugin

        :raises keystoneauth1.exceptions.auth.AuthorizationFailure:
            if a new token fetch fails.
        :raises keystoneauth1.exceptions.auth_plugins.MissingAuthPlugin:
            if a plugin is not available.

        :returns: Authentication headers or None for failure.
        :rtype: :class:`dict`
        """
        auth = self._auth_required(auth, 'fetch a token')
        return auth.get_headers(self)

    def get_token(
        self, auth: ty.Optional['plugin.BaseAuthPlugin'] = None
    ) -> str | None:
        """Return a token as provided by the auth plugin.

        :param auth: The auth plugin to use for token. Overrides the plugin
                     on the session. (optional)
        :type auth: keystoneauth1.plugin.BaseAuthPlugin

        :raises keystoneauth1.exceptions.auth.AuthorizationFailure:
             if a new token fetch fails.
        :raises keystoneauth1.exceptions.auth_plugins.MissingAuthPlugin:
            if a plugin is not available.

        .. warning::
            **DEPRECATED**: This assumes that the only header that is used to
            authenticate a message is ``X-Auth-Token``. This may not be
            correct. Use :meth:`get_auth_headers` instead.

        :returns: A valid token.
        :rtype: string
        """
        return (self.get_auth_headers(auth) or {}).get('X-Auth-Token')

    def get_endpoint(
        self,
        auth: ty.Optional['plugin.BaseAuthPlugin'] = None,
        *,
        endpoint_override: str | None = None,
        **kwargs: ty.Any,
    ) -> str | None:
        """Get an endpoint as provided by the auth plugin.

        :param auth: The auth plugin to use for token. Overrides the plugin on
                     the session. (optional)
        :type auth: keystoneauth1.plugin.BaseAuthPlugin

        :raises keystoneauth1.exceptions.auth_plugins.MissingAuthPlugin:
            if a plugin is not available.

        :returns: An endpoint if available or None.
        :rtype: string
        """
        if endpoint_override:
            return endpoint_override

        auth = self._auth_required(auth, 'determine endpoint URL')

        return auth.get_endpoint(self, **kwargs)

    def get_endpoint_data(
        self,
        auth: ty.Optional['plugin.BaseAuthPlugin'] = None,
        **kwargs: ty.Any,
    ) -> discover.EndpointData | None:
        """Get endpoint data as provided by the auth plugin.

        :param auth: The auth plugin to use for token. Overrides the plugin on
                     the session. (optional)
        :type auth: keystoneauth1.plugin.BaseAuthPlugin

        :raises keystoneauth1.exceptions.auth_plugins.MissingAuthPlugin:
            if a plugin is not available.
        :raises TypeError: If arguments are invalid

        :returns: Endpoint data if available or None.
        :rtype: keystoneauth1.discover.EndpointData
        """
        auth = self._auth_required(auth, 'determine endpoint URL')
        return auth.get_endpoint_data(self, **kwargs)

    def get_api_major_version(
        self,
        auth: ty.Optional['plugin.BaseAuthPlugin'] = None,
        **kwargs: ty.Any,
    ) -> tuple[int | float, ...] | None:
        """Get the major API version as provided by the auth plugin.

        :param auth: The auth plugin to use for token. Overrides the plugin on
                     the session. (optional)
        :type auth: keystoneauth1.plugin.BaseAuthPlugin

        :raises keystoneauth1.exceptions.auth_plugins.MissingAuthPlugin: if a
            plugin is not available.

        :return: The major version of the API of the service discovered.
        :rtype: tuple or None
        """
        auth = self._auth_required(auth, 'determine endpoint URL')
        return auth.get_api_major_version(self, **kwargs)

    def get_all_version_data(
        self,
        auth: ty.Optional['plugin.BaseAuthPlugin'] = None,
        interface: str | list[str] | None = 'public',
        region_name: str | None = None,
        service_type: str | None = None,
    ) -> dict[str, dict[str, dict[str, list[discover.VersionData]]]]:
        """Get version data for all services in the catalog.

        :param auth:
            The auth plugin to use for token. Overrides the plugin on
            the session. (optional)
        :type auth: keystoneauth1.plugin.BaseAuthPlugin
        :param interface:
            Type of endpoint to get version data for. Can be a single value
            or a list of values. A value of None indicates that all interfaces
            should be queried. (optional, defaults to public)
        :param string region_name:
            Region of endpoints to get version data for. A valueof None
            indicates that all regions should be queried. (optional, defaults
            to None)
        :param string service_type:
            Limit the version data to a single service. (optional, defaults
            to None)
        :returns: A dictionary keyed by region_name with values containing
            dictionaries keyed by interface with values being a list of
            `~keystoneauth1.discover.VersionData`.
        """
        auth = self._auth_required(auth, 'determine endpoint URL')
        return auth.get_all_version_data(
            self,
            interface=interface,
            region_name=region_name,
            service_type=service_type,
        )

    def get_auth_connection_params(
        self, auth: ty.Optional['plugin.BaseAuthPlugin'] = None
    ) -> 'plugin.ConnectionParams':
        """Return auth connection params as provided by the auth plugin.

        An auth plugin may specify connection parameters to the request like
        providing a client certificate for communication.

        We restrict the values that may be returned from this function to
        prevent an auth plugin overriding values unrelated to connection
        parameters. The values that are currently accepted are:

        - `cert`: a path to a client certificate, or tuple of client
          certificate and key pair that are used with this request.
        - `verify`: a boolean value to indicate verifying SSL certificates
          against the system CAs or a path to a CA file to verify with.

        These values are passed to the requests library and further information
        on accepted values may be found there.

        :param auth: The auth plugin to use for tokens. Overrides the plugin
                     on the session. (optional)
        :type auth: keystoneauth1.plugin.BaseAuthPlugin

        :raises keystoneauth1.exceptions.auth.AuthorizationFailure:
            if a new token fetch fails.
        :raises keystoneauth1.exceptions.auth_plugins.MissingAuthPlugin:
            if a plugin is not available.
        :raises keystoneauth1.exceptions.auth_plugins.UnsupportedParameters:
            if the plugin returns a parameter that is not supported by this
            session.

        :returns: Authentication headers or None for failure.
        :rtype: :class:`dict`
        """
        auth = self._auth_required(auth, 'fetch connection params')
        params = auth.get_connection_params(self)

        # NOTE(jamielennox): There needs to be some consensus on what
        # parameters are allowed to be modified by the auth plugin here.
        # Ideally I think it would be only the send() parts of the request
        # flow. For now lets just allow certain elements.
        # NOTE(stephenfin): We can't rely on type checking to validate the
        # return values of plugins since these don't happen at runtime. It
        # would be much nicer if we insisted the plugins returned a two-item
        # tuple instead.
        invalid_keys = set(params) - {'cert', 'verify'}
        if invalid_keys:
            raise exceptions.UnsupportedParameters(list(invalid_keys))

        return params

    def invalidate(
        self, auth: ty.Optional['plugin.BaseAuthPlugin'] = None
    ) -> bool:
        """Invalidate an authentication plugin.

        :param auth: The auth plugin to invalidate. Overrides the plugin on the
                     session. (optional)
        :type auth: keystoneauth1.plugin.BaseAuthPlugin
        """
        auth = self._auth_required(auth, 'validate')
        return auth.invalidate()

    def get_user_id(
        self, auth: ty.Optional['plugin.BaseAuthPlugin'] = None
    ) -> str | None:
        """Return the authenticated user_id as provided by the auth plugin.

        :param auth: The auth plugin to use for token. Overrides the plugin
                     on the session. (optional)
        :type auth: keystoneauth1.plugin.BaseAuthPlugin

        :raises keystoneauth1.exceptions.auth.AuthorizationFailure:
            if a new token fetch fails.
        :raises keystoneauth1.exceptions.auth_plugins.MissingAuthPlugin:
            if a plugin is not available.

        :returns: Current user_id or None if not supported by plugin.
        :rtype: :class:`str`
        """
        auth = self._auth_required(auth, 'get user_id')
        return auth.get_user_id(self)

    def get_project_id(
        self, auth: ty.Optional['plugin.BaseAuthPlugin'] = None
    ) -> str | None:
        """Return the authenticated project_id as provided by the auth plugin.

        :param auth: The auth plugin to use for token. Overrides the plugin
                     on the session. (optional)
        :type auth: keystoneauth1.plugin.BaseAuthPlugin

        :raises keystoneauth1.exceptions.auth.AuthorizationFailure:
            if a new token fetch fails.
        :raises keystoneauth1.exceptions.auth_plugins.MissingAuthPlugin:
            if a plugin is not available.

        :returns: Current project_id or None if not supported by plugin.
        :rtype: :class:`str`
        """
        auth = self._auth_required(auth, 'get project_id')
        return auth.get_project_id(self)

    def get_timings(self) -> list[RequestTiming]:
        """Return collected API timing information.

        :returns: List of `RequestTiming` objects.
        """
        return self._api_times

    def reset_timings(self) -> None:
        """Clear API timing information."""
        self._api_times = []


class TCPKeepAliveAdapter(requests.adapters.HTTPAdapter):
    """The custom adapter used to set TCP Keep-Alive on all connections.

    This Adapter also preserves the default behaviour of Requests which
    disables Nagle's Algorithm. See also:
    https://blogs.msdn.com/b/windowsazurestorage/archive/2010/06/25/nagle-s-algorithm-is-not-friendly-towards-small-requests.aspx
    """

    def init_poolmanager(self, *args: ty.Any, **kwargs: ty.Any) -> None:
        if 'socket_options' not in kwargs:
            socket_options = [
                # Keep Nagle's algorithm off
                (socket.IPPROTO_TCP, socket.TCP_NODELAY, 1),
                # Turn on TCP Keep-Alive
                (socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1),
            ]

            # Some operating systems (e.g., OSX) do not support setting
            # keepidle
            if hasattr(socket, 'TCP_KEEPIDLE'):
                socket_options += [
                    # Wait 60 seconds before sending keep-alive probes
                    (socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, 60)
                ]

            # Windows subsystem for Linux does not support this feature
            if (
                hasattr(socket, 'TCP_KEEPCNT')
                and not utils.is_windows_linux_subsystem
            ):
                socket_options += [
                    # Set the maximum number of keep-alive probes
                    (socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 4)
                ]

            if hasattr(socket, 'TCP_KEEPINTVL'):
                socket_options += [
                    # Send keep-alive probes every 15 seconds
                    (socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, 15)
                ]

            # After waiting 60 seconds, and then sending a probe once every 15
            # seconds 4 times, these options should ensure that a connection
            # hands for no longer than 2 minutes before a ConnectionError is
            # raised.
            kwargs['socket_options'] = socket_options
        super().init_poolmanager(*args, **kwargs)  # type: ignore[no-untyped-call]



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import stevedore

from keystoneauth1 import loading
from keystoneauth1.tests.unit.loading import utils


class EntryPointTests(utils.TestCase):
    """Simple test that will check that all entry points are loadable."""

    def test_all_entry_points_are_valid(self):
        errors = []

        def raise_exception_callback(manager, entrypoint, exc):
            error = f"Cannot load '{entrypoint}' entry_point: {exc}'"
            errors.append(error)

        stevedore.ExtensionManager(
            namespace=loading.PLUGIN_NAMESPACE,
            on_load_failure_callback=raise_exception_callback,
        )

        self.assertEqual([], errors)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import typing as ty

from keystoneauth1 import plugin

if ty.TYPE_CHECKING:
    from keystoneauth1 import session as ks_session


class NoAuth(plugin.FixedEndpointPlugin):
    """A provider that will always use no auth.

    This is useful to unify session/adapter loading for services
    that might be deployed in standalone/noauth mode.
    """

    def get_token(self, session: 'ks_session.Session') -> str | None:
        return 'notused'



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from keystoneauth1.extras import kerberos
from keystoneauth1 import fixture as ks_fixture
from keystoneauth1 import session
from keystoneauth1.tests.unit.extras.kerberos import base


class TestMappedAuth(base.TestCase):
    def setUp(self):
        if kerberos.requests_kerberos is None:
            self.skipTest("Kerberos support isn't available.")

        super().setUp()

        self.protocol = uuid.uuid4().hex
        self.identity_provider = uuid.uuid4().hex

    @property
    def token_url(self):
        fmt = '%s/OS-FEDERATION/identity_providers/%s/protocols/%s/auth'
        return fmt % (self.TEST_V3_URL, self.identity_provider, self.protocol)

    def test_unscoped_mapped_auth(self):
        token_id, _ = self.kerberos_mock.mock_auth_success(
            url=self.token_url, method='GET'
        )

        plugin = kerberos.MappedKerberos(
            auth_url=self.TEST_V3_URL,
            protocol=self.protocol,
            identity_provider=self.identity_provider,
        )

        sess = session.Session()
        tok = plugin.get_token(sess)

        self.assertEqual(token_id, tok)

    def test_project_scoped_mapped_auth(self):
        self.kerberos_mock.mock_auth_success(url=self.token_url, method='GET')

        scoped_id = uuid.uuid4().hex
        scoped_body = ks_fixture.V3Token()
        scoped_body.set_project_scope()

        self.requests_mock.post(
            f'{self.TEST_V3_URL}/auth/tokens',
            json=scoped_body,
            headers={
                'X-Subject-Token': scoped_id,
                'Content-Type': 'application/json',
            },
        )

        plugin = kerberos.MappedKerberos(
            auth_url=self.TEST_V3_URL,
            protocol=self.protocol,
            identity_provider=self.identity_provider,
            project_id=scoped_body.project_id,
        )

        sess = session.Session()
        tok = plugin.get_token(sess)
        proj = plugin.get_project_id(sess)

        self.assertEqual(scoped_id, tok)
        self.assertEqual(scoped_body.project_id, proj)

    def test_authenticate_with_mutual_authentication_required(self):
        self.kerberos_mock.mock_auth_success(url=self.token_url, method='GET')

        scoped_id = uuid.uuid4().hex
        scoped_body = ks_fixture.V3Token()
        scoped_body.set_project_scope()

        self.requests_mock.post(
            f'{self.TEST_V3_URL}/auth/tokens',
            json=scoped_body,
            headers={
                'X-Subject-Token': scoped_id,
                'Content-Type': 'application/json',
            },
        )

        plugin = kerberos.MappedKerberos(
            auth_url=self.TEST_V3_URL,
            protocol=self.protocol,
            identity_provider=self.identity_provider,
            project_id=scoped_body.project_id,
            mutual_auth='required',
        )

        sess = session.Session()
        tok = plugin.get_token(sess)
        proj = plugin.get_project_id(sess)

        self.assertEqual(scoped_id, tok)
        self.assertEqual(scoped_body.project_id, proj)
        self.assertEqual(self.kerberos_mock.called_auth_server, True)

    def test_authenticate_with_mutual_authentication_disabled(self):
        self.kerberos_mock.mock_auth_success(url=self.token_url, method='GET')

        scoped_id = uuid.uuid4().hex
        scoped_body = ks_fixture.V3Token()
        scoped_body.set_project_scope()

        self.requests_mock.post(
            f'{self.TEST_V3_URL}/auth/tokens',
            json=scoped_body,
            headers={
                'X-Subject-Token': scoped_id,
                'Content-Type': 'application/json',
            },
        )

        plugin = kerberos.MappedKerberos(
            auth_url=self.TEST_V3_URL,
            protocol=self.protocol,
            identity_provider=self.identity_provider,
            project_id=scoped_body.project_id,
            mutual_auth='disabled',
        )

        sess = session.Session()
        tok = plugin.get_token(sess)
        proj = plugin.get_project_id(sess)

        self.assertEqual(scoped_id, tok)
        self.assertEqual(scoped_body.project_id, proj)
        self.assertEqual(self.kerberos_mock.called_auth_server, False)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import typing as ty
import uuid

from keystoneauth1 import _utils
from keystoneauth1.fixture import exception


class _Service(dict[str, ty.Any]):
    """One of the services that exist in the catalog.

    You use this by adding a service to a token which returns an instance of
    this object and then you can add_endpoints to the service.
    """

    def add_endpoint(self, interface, url, region=None, id=None):
        data = {
            'id': id or uuid.uuid4().hex,
            'interface': interface,
            'url': url,
            'region': region,
            'region_id': region,
        }
        self.setdefault('endpoints', []).append(data)
        return data

    def add_standard_endpoints(
        self, public=None, admin=None, internal=None, region=None
    ):
        ret = []

        if public:
            ret.append(self.add_endpoint('public', public, region=region))
        if admin:
            ret.append(self.add_endpoint('admin', admin, region=region))
        if internal:
            ret.append(self.add_endpoint('internal', internal, region=region))

        return ret


class Token(dict[str, ty.Any]):
    """A V3 Keystone token that can be used for testing.

    This object is designed to allow clients to generate a correct V3 token for
    use in there test code. It should prevent clients from having to know the
    correct token format and allow them to test the portions of token handling
    that matter to them and not copy and paste sample.
    """

    def __init__(
        self,
        expires=None,
        issued=None,
        user_id=None,
        user_name=None,
        user_domain_id=None,
        user_domain_name=None,
        methods=None,
        project_id=None,
        project_name=None,
        project_domain_id=None,
        project_domain_name=None,
        domain_id=None,
        domain_name=None,
        trust_id=None,
        trust_impersonation=None,
        trustee_user_id=None,
        trustor_user_id=None,
        application_credential_id=None,
        application_credential_access_rules=None,
        oauth_access_token_id=None,
        oauth_consumer_id=None,
        audit_id=None,
        audit_chain_id=None,
        is_admin_project=None,
        project_is_domain=None,
        oauth2_thumbprint=None,
    ):
        super().__init__()

        self.user_id = user_id or uuid.uuid4().hex
        self.user_name = user_name or uuid.uuid4().hex
        self.user_domain_id = user_domain_id or uuid.uuid4().hex
        self.user_domain_name = user_domain_name or uuid.uuid4().hex
        self.audit_id = audit_id or uuid.uuid4().hex

        if not methods:
            methods = ['password']
        self.methods.extend(methods)

        if not issued:
            issued = _utils.before_utcnow(minutes=2)

        try:
            self.issued = issued
        except (TypeError, AttributeError):
            # issued should be able to be passed as a string so ignore
            self.issued_str = issued

        if not expires:
            expires = self.issued + datetime.timedelta(hours=1)

        try:
            self.expires = expires
        except (TypeError, AttributeError):
            # expires should be able to be passed as a string so ignore
            self.expires_str = expires

        if (
            project_id
            or project_name
            or project_domain_id
            or project_domain_name
        ):
            self.set_project_scope(
                id=project_id,
                name=project_name,
                domain_id=project_domain_id,
                domain_name=project_domain_name,
                is_domain=project_is_domain,
            )

        if domain_id or domain_name:
            self.set_domain_scope(id=domain_id, name=domain_name)

        if (
            trust_id
            or (trust_impersonation is not None)
            or trustee_user_id
            or trustor_user_id
        ):
            self.set_trust_scope(
                id=trust_id,
                impersonation=trust_impersonation,
                trustee_user_id=trustee_user_id,
                trustor_user_id=trustor_user_id,
            )

        if application_credential_id:
            self.set_application_credential(
                application_credential_id,
                access_rules=application_credential_access_rules,
            )

        if oauth_access_token_id or oauth_consumer_id:
            self.set_oauth(
                access_token_id=oauth_access_token_id,
                consumer_id=oauth_consumer_id,
            )

        if audit_chain_id:
            self.audit_chain_id = audit_chain_id

        if is_admin_project is not None:
            self.is_admin_project = is_admin_project

        if oauth2_thumbprint:
            self.oauth2_thumbprint = oauth2_thumbprint

    @property
    def root(self):
        return self.setdefault('token', {})

    @property
    def expires_str(self):
        return self.root.get('expires_at')

    @expires_str.setter
    def expires_str(self, value):
        self.root['expires_at'] = value

    @property
    def expires(self):
        return _utils.parse_isotime(self.expires_str)

    @expires.setter
    def expires(self, value):
        self.expires_str = value.isoformat()

    @property
    def issued_str(self):
        return self.root.get('issued_at')

    @issued_str.setter
    def issued_str(self, value):
        self.root['issued_at'] = value

    @property
    def issued(self):
        return _utils.parse_isotime(self.issued_str)

    @issued.setter
    def issued(self, value):
        self.issued_str = value.isoformat()

    @property
    def _user(self):
        return self.root.setdefault('user', {})

    @property
    def user_id(self):
        return self._user.get('id')

    @user_id.setter
    def user_id(self, value):
        self._user['id'] = value

    @property
    def user_name(self):
        return self._user.get('name')

    @user_name.setter
    def user_name(self, value):
        self._user['name'] = value

    @property
    def _user_domain(self):
        return self._user.setdefault('domain', {})

    @_user_domain.setter
    def _user_domain(self, domain):
        self._user['domain'] = domain

    @property
    def user_domain_id(self):
        return self._user_domain.get('id')

    @user_domain_id.setter
    def user_domain_id(self, value):
        self._user_domain['id'] = value

    @property
    def user_domain_name(self):
        return self._user_domain.get('name')

    @user_domain_name.setter
    def user_domain_name(self, value):
        self._user_domain['name'] = value

    @property
    def methods(self):
        return self.root.setdefault('methods', [])

    @property
    def project_id(self):
        return self.root.get('project', {}).get('id')

    @project_id.setter
    def project_id(self, value):
        self.root.setdefault('project', {})['id'] = value

    @property
    def project_is_domain(self):
        return self.root.get('is_domain')

    @project_is_domain.setter
    def project_is_domain(self, value):
        self.root['is_domain'] = value

    @property
    def project_name(self):
        return self.root.get('project', {}).get('name')

    @project_name.setter
    def project_name(self, value):
        self.root.setdefault('project', {})['name'] = value

    @property
    def project_domain_id(self):
        return self.root.get('project', {}).get('domain', {}).get('id')

    @project_domain_id.setter
    def project_domain_id(self, value):
        project = self.root.setdefault('project', {})
        project.setdefault('domain', {})['id'] = value

    @property
    def project_domain_name(self):
        return self.root.get('project', {}).get('domain', {}).get('name')

    @project_domain_name.setter
    def project_domain_name(self, value):
        project = self.root.setdefault('project', {})
        project.setdefault('domain', {})['name'] = value

    @property
    def domain_id(self):
        return self.root.get('domain', {}).get('id')

    @domain_id.setter
    def domain_id(self, value):
        self.root.setdefault('domain', {})['id'] = value

    @property
    def domain_name(self):
        return self.root.get('domain', {}).get('name')

    @domain_name.setter
    def domain_name(self, value):
        self.root.setdefault('domain', {})['name'] = value

    @property
    def system(self):
        return self.root.get('system', {})

    @system.setter
    def system(self, value):
        return self.root.setdefault('system', value)

    @property
    def trust_id(self):
        return self.root.get('OS-TRUST:trust', {}).get('id')

    @trust_id.setter
    def trust_id(self, value):
        self.root.setdefault('OS-TRUST:trust', {})['id'] = value

    @property
    def trust_impersonation(self):
        return self.root.get('OS-TRUST:trust', {}).get('impersonation')

    @trust_impersonation.setter
    def trust_impersonation(self, value):
        self.root.setdefault('OS-TRUST:trust', {})['impersonation'] = value

    @property
    def trustee_user_id(self):
        trust = self.root.get('OS-TRUST:trust', {})
        return trust.get('trustee_user', {}).get('id')

    @trustee_user_id.setter
    def trustee_user_id(self, value):
        trust = self.root.setdefault('OS-TRUST:trust', {})
        trust.setdefault('trustee_user', {})['id'] = value

    @property
    def trustor_user_id(self):
        trust = self.root.get('OS-TRUST:trust', {})
        return trust.get('trustor_user', {}).get('id')

    @trustor_user_id.setter
    def trustor_user_id(self, value):
        trust = self.root.setdefault('OS-TRUST:trust', {})
        trust.setdefault('trustor_user', {})['id'] = value

    @property
    def application_credential_id(self):
        return self.root.get('application_credential', {}).get('id')

    @application_credential_id.setter
    def application_credential_id(self, value):
        application_credential = self.root.setdefault(
            'application_credential', {}
        )
        application_credential.setdefault('id', value)

    @property
    def application_credential_access_rules(self):
        return self.root.get('application_credential', {}).get('access_rules')

    @application_credential_access_rules.setter
    def application_credential_access_rules(self, value):
        application_credential = self.root.setdefault(
            'application_credential', {}
        )
        application_credential.setdefault('access_rules', value)

    @property
    def oauth_access_token_id(self):
        return self.root.get('OS-OAUTH1', {}).get('access_token_id')

    @oauth_access_token_id.setter
    def oauth_access_token_id(self, value):
        self.root.setdefault('OS-OAUTH1', {})['access_token_id'] = value

    @property
    def oauth_consumer_id(self):
        return self.root.get('OS-OAUTH1', {}).get('consumer_id')

    @oauth_consumer_id.setter
    def oauth_consumer_id(self, value):
        self.root.setdefault('OS-OAUTH1', {})['consumer_id'] = value

    @property
    def audit_id(self):
        try:
            return self.root.get('audit_ids', [])[0]
        except IndexError:
            return None

    @audit_id.setter
    def audit_id(self, value):
        audit_chain_id = self.audit_chain_id
        lval = [value] if audit_chain_id else [value, audit_chain_id]
        self.root['audit_ids'] = lval

    @property
    def audit_chain_id(self):
        try:
            return self.root.get('audit_ids', [])[1]
        except IndexError:
            return None

    @audit_chain_id.setter
    def audit_chain_id(self, value):
        self.root['audit_ids'] = [self.audit_id, value]

    @property
    def role_ids(self):
        return [r['id'] for r in self.root.get('roles', [])]

    @property
    def role_names(self):
        return [r['name'] for r in self.root.get('roles', [])]

    @property
    def is_admin_project(self):
        return self.root.get('is_admin_project')

    @is_admin_project.setter
    def is_admin_project(self, value):
        self.root['is_admin_project'] = value

    @is_admin_project.deleter
    def is_admin_project(self):
        self.root.pop('is_admin_project', None)

    @property
    def oauth2_thumbprint(self):
        return self.root.get('oauth2_credential', {}).get('x5t#S256')

    @oauth2_thumbprint.setter
    def oauth2_thumbprint(self, value):
        self.root.setdefault('oauth2_credential', {})['x5t#S256'] = value

    @property
    def oauth2_credential(self):
        return self.root.get('oauth2_credential')

    def validate(self):
        project = self.root.get('project')
        domain = self.root.get('domain')
        system = self.root.get('system')
        trust = self.root.get('OS-TRUST:trust')
        catalog = self.root.get('catalog')
        roles = self.root.get('roles')
        scoped = project or domain or trust

        if sum((bool(project), bool(domain), bool(trust), bool(system))) > 1:
            msg = 'You cannot scope to multiple targets'
            raise exception.FixtureValidationError(msg)

        if catalog and not scoped:
            msg = 'You cannot have a service catalog on an unscoped token'
            raise exception.FixtureValidationError(msg)

        if scoped and not self._user.get('roles'):
            msg = 'You must have roles on a token to scope it'
            raise exception.FixtureValidationError(msg)

        if bool(scoped) != bool(roles):
            msg = 'You must be scoped to have roles and vice-versa'
            raise exception.FixtureValidationError(msg)

    def add_role(self, name=None, id=None):
        roles = self.root.setdefault('roles', [])
        data = {'id': id or uuid.uuid4().hex, 'name': name or uuid.uuid4().hex}
        roles.append(data)
        return data

    def add_service(self, type, name=None, id=None):
        service = _Service(type=type, id=id or uuid.uuid4().hex)
        if name:
            service['name'] = name
        self.root.setdefault('catalog', []).append(service)
        return service

    def remove_service(self, type):
        self.root.setdefault('catalog', [])
        self.root['catalog'] = [
            f for f in self.root.setdefault('catalog', []) if f['type'] != type
        ]

    def set_project_scope(
        self,
        id=None,
        name=None,
        domain_id=None,
        domain_name=None,
        is_domain=None,
    ):
        self.project_id = id or uuid.uuid4().hex
        self.project_name = name or uuid.uuid4().hex
        self.project_domain_id = domain_id or uuid.uuid4().hex
        self.project_domain_name = domain_name or uuid.uuid4().hex

        if is_domain is not None:
            self.project_is_domain = is_domain

    def set_domain_scope(self, id=None, name=None):
        self.domain_id = id or uuid.uuid4().hex
        self.domain_name = name or uuid.uuid4().hex

    def set_system_scope(self):
        # NOTE(lbragstad): In the future it might be possible to scope a token
        # to a subset of the entire system (e.g. a specific service, region, or
        # service within a region). Until then, the only system scope is the
        # entire system.
        self.system = {'all': True}

    def set_trust_scope(
        self,
        id=None,
        impersonation=False,
        trustee_user_id=None,
        trustor_user_id=None,
    ):
        self.trust_id = id or uuid.uuid4().hex
        self.trust_impersonation = impersonation
        self.trustee_user_id = trustee_user_id or uuid.uuid4().hex
        self.trustor_user_id = trustor_user_id or uuid.uuid4().hex

    def set_oauth(self, access_token_id=None, consumer_id=None):
        self.oauth_access_token_id = access_token_id or uuid.uuid4().hex
        self.oauth_consumer_id = consumer_id or uuid.uuid4().hex

    def set_application_credential(
        self, application_credential_id, access_rules=None
    ):
        self.application_credential_id = application_credential_id
        if access_rules is not None:
            self.application_credential_access_rules = access_rules

    @property
    def service_providers(self):
        return self.root.get('service_providers')

    def add_service_provider(self, sp_id, sp_auth_url, sp_url):
        _service_providers = self.root.setdefault('service_providers', [])
        sp = {'id': sp_id, 'auth_url': sp_auth_url, 'sp_url': sp_url}
        _service_providers.append(sp)
        return sp

    def set_bind(self, name, data):
        self.root.setdefault('bind', {})[name] = data


class V3FederationToken(Token):
    """A V3 Keystone Federation token that can be used for testing.

    Similar to V3Token, this object is designed to allow clients to generate
    a correct V3 federation token for use in test code.
    """

    FEDERATED_DOMAIN_ID = 'Federated'

    def __init__(
        self, methods=None, identity_provider=None, protocol=None, groups=None
    ):
        methods = methods or ['saml2']
        super().__init__(methods=methods)
        self._user_domain = {'id': V3FederationToken.FEDERATED_DOMAIN_ID}
        self.add_federation_info_to_user(identity_provider, protocol, groups)

    def add_federation_info_to_user(
        self, identity_provider=None, protocol=None, groups=None
    ):
        data = {
            "OS-FEDERATION": {
                "identity_provider": identity_provider or uuid.uuid4().hex,
                "protocol": protocol or uuid.uuid4().hex,
                "groups": groups or [{"id": uuid.uuid4().hex}],
            }
        }
        self._user.update(data)
        return data



# Copyright 2010-2011 OpenStack Foundation
# Copyright (c) 2013 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystoneauth1.tests.unit.extras.kerberos import utils
from keystoneauth1.tests.unit import utils as test_utils


REQUEST = {'auth': {'identity': {'methods': ['kerberos'], 'kerberos': {}}}}


class TestCase(test_utils.TestCase):
    """Test case base class for Kerberos unit tests."""

    TEST_V3_URL = test_utils.TestCase.TEST_ROOT_URL + 'v3'

    def setUp(self):
        super().setUp()

        km = utils.KerberosMock(self.requests_mock)
        self.kerberos_mock = self.useFixture(km)

    def assertRequestBody(self, body=None):
        """Ensure the request body is the standard Kerberos auth request.

        :param dict body: the body to compare. If not provided the last request
                          body will be used.
        """
        if not body:
            body = self.requests_mock.last_request.json()

        self.assertEqual(REQUEST, body)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import json
import time
import uuid

from keystoneauth1 import _utils as ksa_utils
from keystoneauth1 import access
from keystoneauth1 import exceptions
from keystoneauth1 import fixture
from keystoneauth1.identity import v2
from keystoneauth1 import session
from keystoneauth1.tests.unit import utils


class V2IdentityPlugin(utils.TestCase):
    TEST_ROOT_URL = 'http://127.0.0.1:5000/'
    TEST_URL = '{}{}'.format(TEST_ROOT_URL, 'v2.0')
    TEST_ROOT_ADMIN_URL = 'http://127.0.0.1:35357/'
    TEST_ADMIN_URL = '{}{}'.format(TEST_ROOT_ADMIN_URL, 'v2.0')

    TEST_PASS = 'password'

    TEST_SERVICE_CATALOG = [
        {
            "endpoints": [
                {
                    "adminURL": "http://cdn.admin-nets.local:8774/v1.0",
                    "region": "RegionOne",
                    "internalURL": "http://127.0.0.1:8774/v1.0",
                    "publicURL": "http://cdn.admin-nets.local:8774/v1.0/",
                }
            ],
            "type": "nova_compat",
            "name": "nova_compat",
        },
        {
            "endpoints": [
                {
                    "adminURL": "http://nova/novapi/admin",
                    "region": "RegionOne",
                    "internalURL": "http://nova/novapi/internal",
                    "publicURL": "http://nova/novapi/public",
                }
            ],
            "type": "compute",
            "name": "nova",
        },
        {
            "endpoints": [
                {
                    "adminURL": "http://glance/glanceapi/admin",
                    "region": "RegionOne",
                    "internalURL": "http://glance/glanceapi/internal",
                    "publicURL": "http://glance/glanceapi/public",
                }
            ],
            "type": "image",
            "name": "glance",
        },
        {
            "endpoints": [
                {
                    "adminURL": TEST_ADMIN_URL,
                    "region": "RegionOne",
                    "internalURL": "http://127.0.0.1:5000/v2.0",
                    "publicURL": "http://127.0.0.1:5000/v2.0",
                }
            ],
            "type": "identity",
            "name": "keystone",
        },
        {
            "endpoints": [
                {
                    "adminURL": "http://swift/swiftapi/admin",
                    "region": "RegionOne",
                    "internalURL": "http://swift/swiftapi/internal",
                    "publicURL": "http://swift/swiftapi/public",
                }
            ],
            "type": "object-store",
            "name": "swift",
        },
    ]

    def setUp(self):
        super().setUp()

        nextyear = 1 + time.gmtime().tm_year
        self.TEST_RESPONSE_DICT = {
            "access": {
                "token": {
                    "expires": f"{nextyear}-02-01T00:00:10.000123Z",
                    "id": self.TEST_TOKEN,
                    "tenant": {"id": self.TEST_TENANT_ID},
                },
                "user": {"id": self.TEST_USER},
                "serviceCatalog": self.TEST_SERVICE_CATALOG,
            }
        }

    def stub_auth(self, **kwargs):
        self.stub_url('POST', ['tokens'], **kwargs)

    def test_authenticate_with_username_password(self):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)
        a = v2.Password(
            self.TEST_URL, username=self.TEST_USER, password=self.TEST_PASS
        )
        self.assertIsNone(a.user_id)
        self.assertFalse(a.has_scope_parameters)
        s = session.Session(a)
        self.assertEqual(
            {'X-Auth-Token': self.TEST_TOKEN}, s.get_auth_headers()
        )

        req = {
            'auth': {
                'passwordCredentials': {
                    'username': self.TEST_USER,
                    'password': self.TEST_PASS,
                }
            }
        }
        self.assertRequestBodyIs(json=req)
        self.assertRequestHeaderEqual('Content-Type', 'application/json')
        self.assertRequestHeaderEqual('Accept', 'application/json')
        self.assertEqual(s.auth.auth_ref.auth_token, self.TEST_TOKEN)

    def test_authenticate_with_user_id_password(self):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)
        a = v2.Password(
            self.TEST_URL, user_id=self.TEST_USER, password=self.TEST_PASS
        )
        self.assertIsNone(a.username)
        self.assertFalse(a.has_scope_parameters)
        s = session.Session(a)
        self.assertEqual(
            {'X-Auth-Token': self.TEST_TOKEN}, s.get_auth_headers()
        )

        req = {
            'auth': {
                'passwordCredentials': {
                    'userId': self.TEST_USER,
                    'password': self.TEST_PASS,
                }
            }
        }
        self.assertRequestBodyIs(json=req)
        self.assertRequestHeaderEqual('Content-Type', 'application/json')
        self.assertRequestHeaderEqual('Accept', 'application/json')
        self.assertEqual(s.auth.auth_ref.auth_token, self.TEST_TOKEN)

    def test_authenticate_with_username_password_scoped(self):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)
        a = v2.Password(
            self.TEST_URL,
            username=self.TEST_USER,
            password=self.TEST_PASS,
            tenant_id=self.TEST_TENANT_ID,
        )
        self.assertTrue(a.has_scope_parameters)
        self.assertIsNone(a.user_id)
        s = session.Session(a)
        self.assertEqual(
            {'X-Auth-Token': self.TEST_TOKEN}, s.get_auth_headers()
        )

        req = {
            'auth': {
                'passwordCredentials': {
                    'username': self.TEST_USER,
                    'password': self.TEST_PASS,
                },
                'tenantId': self.TEST_TENANT_ID,
            }
        }
        self.assertRequestBodyIs(json=req)
        self.assertEqual(s.auth.auth_ref.auth_token, self.TEST_TOKEN)

    def test_authenticate_with_user_id_password_scoped(self):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)
        a = v2.Password(
            self.TEST_URL,
            user_id=self.TEST_USER,
            password=self.TEST_PASS,
            tenant_id=self.TEST_TENANT_ID,
        )
        self.assertIsNone(a.username)
        self.assertTrue(a.has_scope_parameters)
        s = session.Session(a)
        self.assertEqual(
            {'X-Auth-Token': self.TEST_TOKEN}, s.get_auth_headers()
        )

        req = {
            'auth': {
                'passwordCredentials': {
                    'userId': self.TEST_USER,
                    'password': self.TEST_PASS,
                },
                'tenantId': self.TEST_TENANT_ID,
            }
        }
        self.assertRequestBodyIs(json=req)
        self.assertEqual(s.auth.auth_ref.auth_token, self.TEST_TOKEN)

    def test_authenticate_with_token(self):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)
        a = v2.Token(self.TEST_URL, 'foo')
        s = session.Session(a)
        self.assertEqual(
            {'X-Auth-Token': self.TEST_TOKEN}, s.get_auth_headers()
        )

        req = {'auth': {'token': {'id': 'foo'}}}
        self.assertRequestBodyIs(json=req)
        self.assertRequestHeaderEqual('x-Auth-Token', 'foo')
        self.assertRequestHeaderEqual('Content-Type', 'application/json')
        self.assertRequestHeaderEqual('Accept', 'application/json')
        self.assertEqual(s.auth.auth_ref.auth_token, self.TEST_TOKEN)

    def test_with_trust_id(self):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)
        a = v2.Password(
            self.TEST_URL,
            username=self.TEST_USER,
            password=self.TEST_PASS,
            trust_id='trust',
        )
        self.assertTrue(a.has_scope_parameters)
        s = session.Session(a)
        self.assertEqual(
            {'X-Auth-Token': self.TEST_TOKEN}, s.get_auth_headers()
        )

        req = {
            'auth': {
                'passwordCredentials': {
                    'username': self.TEST_USER,
                    'password': self.TEST_PASS,
                },
                'trust_id': 'trust',
            }
        }

        self.assertRequestBodyIs(json=req)
        self.assertEqual(s.auth.auth_ref.auth_token, self.TEST_TOKEN)

    def _do_service_url_test(self, base_url, endpoint_filter):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)
        self.stub_url(
            'GET', ['path'], base_url=base_url, text='SUCCESS', status_code=200
        )

        a = v2.Password(
            self.TEST_URL, username=self.TEST_USER, password=self.TEST_PASS
        )
        s = session.Session(auth=a)

        resp = s.get('/path', endpoint_filter=endpoint_filter)

        self.assertEqual(resp.status_code, 200)
        self.assertEqual(
            self.requests_mock.last_request.url, base_url + '/path'
        )

    def test_service_url(self):
        endpoint_filter = {
            'service_type': 'compute',
            'interface': 'admin',
            'service_name': 'nova',
        }
        self._do_service_url_test('http://nova/novapi/admin', endpoint_filter)

    def test_service_url_defaults_to_public(self):
        endpoint_filter = {'service_type': 'compute'}
        self._do_service_url_test('http://nova/novapi/public', endpoint_filter)

    def test_endpoint_filter_without_service_type_fails(self):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)

        a = v2.Password(
            self.TEST_URL, username=self.TEST_USER, password=self.TEST_PASS
        )
        s = session.Session(auth=a)

        self.assertRaises(
            exceptions.EndpointNotFound,
            s.get,
            '/path',
            endpoint_filter={'interface': 'admin'},
        )

    def test_full_url_overrides_endpoint_filter(self):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)
        self.stub_url(
            'GET',
            [],
            base_url='http://testurl/',
            text='SUCCESS',
            status_code=200,
        )

        a = v2.Password(
            self.TEST_URL, username=self.TEST_USER, password=self.TEST_PASS
        )
        s = session.Session(auth=a)

        resp = s.get(
            'http://testurl/', endpoint_filter={'service_type': 'compute'}
        )
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.text, 'SUCCESS')

    def test_invalid_auth_response_dict(self):
        self.stub_auth(json={'hello': 'world'})

        a = v2.Password(
            self.TEST_URL, username=self.TEST_USER, password=self.TEST_PASS
        )
        s = session.Session(auth=a)

        self.assertRaises(
            exceptions.InvalidResponse, s.get, 'http://any', authenticated=True
        )

    def test_invalid_auth_response_type(self):
        self.stub_url('POST', ['tokens'], text='testdata')

        a = v2.Password(
            self.TEST_URL, username=self.TEST_USER, password=self.TEST_PASS
        )
        s = session.Session(auth=a)

        self.assertRaises(
            exceptions.InvalidResponse, s.get, 'http://any', authenticated=True
        )

    def test_invalidate_response(self):
        resp_data1 = copy.deepcopy(self.TEST_RESPONSE_DICT)
        resp_data2 = copy.deepcopy(self.TEST_RESPONSE_DICT)

        resp_data1['access']['token']['id'] = 'token1'
        resp_data2['access']['token']['id'] = 'token2'

        auth_responses = [{'json': resp_data1}, {'json': resp_data2}]
        self.stub_auth(response_list=auth_responses)

        a = v2.Password(
            self.TEST_URL, username=self.TEST_USER, password=self.TEST_PASS
        )
        s = session.Session(auth=a)

        self.assertEqual('token1', s.get_token())
        self.assertEqual({'X-Auth-Token': 'token1'}, s.get_auth_headers())

        a.invalidate()
        self.assertEqual('token2', s.get_token())
        self.assertEqual({'X-Auth-Token': 'token2'}, s.get_auth_headers())

    def test_doesnt_log_password(self):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)
        password = uuid.uuid4().hex

        a = v2.Password(
            self.TEST_URL, username=self.TEST_USER, password=password
        )
        s = session.Session(auth=a)
        self.assertEqual(self.TEST_TOKEN, s.get_token())
        self.assertEqual(
            {'X-Auth-Token': self.TEST_TOKEN}, s.get_auth_headers()
        )
        self.assertNotIn(password, self.logger.output)

    def test_password_with_no_user_id_or_name(self):
        self.assertRaises(
            TypeError, v2.Password, self.TEST_URL, password=self.TEST_PASS
        )

    def test_password_cache_id(self):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)
        trust_id = uuid.uuid4().hex

        a = v2.Password(
            self.TEST_URL,
            username=self.TEST_USER,
            password=self.TEST_PASS,
            trust_id=trust_id,
        )

        b = v2.Password(
            self.TEST_URL,
            username=self.TEST_USER,
            password=self.TEST_PASS,
            trust_id=trust_id,
        )

        a_id = a.get_cache_id()
        b_id = b.get_cache_id()

        self.assertEqual(a_id, b_id)

        c = v2.Password(
            self.TEST_URL,
            username=self.TEST_USER,
            password=self.TEST_PASS,
            tenant_id=trust_id,
        )  # same value different param

        c_id = c.get_cache_id()

        self.assertNotEqual(a_id, c_id)

        self.assertIsNone(a.get_auth_state())
        self.assertIsNone(b.get_auth_state())
        self.assertIsNone(c.get_auth_state())

        s = session.Session()
        self.assertEqual(self.TEST_TOKEN, a.get_token(s))
        self.assertTrue(self.requests_mock.called)

    def test_password_change_auth_state(self):
        self.stub_auth(json=self.TEST_RESPONSE_DICT)

        expired = ksa_utils.before_utcnow(days=2)
        token = fixture.V2Token(expires=expired)

        auth_ref = access.create(body=token)

        a = v2.Password(
            self.TEST_URL,
            username=self.TEST_USER,
            password=self.TEST_PASS,
            tenant_id=uuid.uuid4().hex,
        )

        initial_cache_id = a.get_cache_id()

        state = a.get_auth_state()
        self.assertIsNone(state)

        state = json.dumps(
            {'auth_token': auth_ref.auth_token, 'body': auth_ref._data}
        )
        a.set_auth_state(state)

        self.assertEqual(token.token_id, a.auth_ref.auth_token)

        s = session.Session()
        self.assertEqual(self.TEST_TOKEN, a.get_token(s))  # updates expired
        self.assertEqual(initial_cache_id, a.get_cache_id())



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import typing as ty

from keystoneauth1.exceptions import base

if ty.TYPE_CHECKING:
    from keystoneauth1 import loading


__all__ = (
    'AuthPluginException',
    'MissingAuthPlugin',
    'NoMatchingPlugin',
    'UnsupportedParameters',
    'OptionError',
    'MissingRequiredOptions',
)


class AuthPluginException(base.ClientException):
    message = "Unknown error with authentication plugins."


class MissingAuthPlugin(AuthPluginException):
    message = "An authenticated request is required but no plugin available."


class NoMatchingPlugin(AuthPluginException):
    """No auth plugins could be created from the parameters provided.

    :param str name: The name of the plugin that was attempted to load.

    .. py:attribute:: name

        The name of the plugin that was attempted to load.
    """

    def __init__(self, name: str):
        self.name = name
        msg = f'The plugin {name} could not be found'
        super().__init__(msg)


class UnsupportedParameters(AuthPluginException):
    """A parameter that was provided or returned is not supported.

    :param list(str) names: Names of the unsupported parameters.

    .. py:attribute:: names

        Names of the unsupported parameters.
    """

    def __init__(self, names: ty.Sequence[str]):
        self.names = names

        m = 'The following parameters were given that are unsupported: %s'
        super().__init__(m % ', '.join(self.names))


class OptionError(AuthPluginException):
    """A requirement of this plugin loader was not met.

    This error can be raised by a specific plugin loader during the
    load_from_options stage to indicate a parameter problem that can not be
    handled by the generic options loader.

    The intention here is that a plugin can do checks like if a name parameter
    is provided then a domain parameter must also be provided, but that Opt
    checking doesn't handle.
    """


class MissingRequiredOptions(OptionError):
    """One or more required options were not provided.

    :param list(keystoneauth1.loading.Opt) options: Missing options.

    .. py:attribute:: options

        List of the missing options.
    """

    def __init__(self, options: ty.Sequence['loading.Opt']):
        self.options = options

        names = ", ".join(o.dest for o in options)
        m = 'Auth plugin requires parameters which were not given: %s'
        super().__init__(m % names)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

from lxml import etree

from keystoneauth1 import session
from keystoneauth1.tests.unit import utils

ROOTDIR = os.path.dirname(os.path.abspath(__file__))
XMLDIR = os.path.join(ROOTDIR, 'examples', 'xml/')


def make_oneline(s):
    return etree.tostring(etree.XML(s)).replace(b'\n', b'')


def _load_xml(filename):
    with open(XMLDIR + filename, 'rb') as f:
        return f.read()


class TestCase(utils.TestCase):
    TEST_URL = 'https://keystone:5000/v3'

    def setUp(self):
        super().setUp()
        self.session = session.Session()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
from unittest import mock

import betamax
from requests import models
import testtools

try:
    from requests.packages.urllib3._collections import HTTPHeaderDict
except ImportError:
    from betamax.headers import HTTPHeaderDict

from keystoneauth1.fixture import hooks


class TestBetamaxHooks(testtools.TestCase):
    def test_pre_record_hook_v3(self):
        fixtures_path = 'keystoneauth1/tests/unit/data'

        with betamax.Betamax.configure() as config:
            config.before_record(callback=hooks.pre_record_hook)

        cassette = betamax.cassette.Cassette(
            'test_pre_record_hook',
            'json',
            record_mode=None,
            cassette_library_dir=fixtures_path,
        )

        # Create a new object to serialize
        r = models.Response()
        r.status_code = 200
        r.reason = 'OK'
        r.encoding = 'utf-8'
        r.headers = {}
        r.url = 'http://localhost:35357/'

        # load request and response
        with open(f'{fixtures_path}/keystone_v3_sample_response.json') as f:
            response_content = json.loads(f.read())
        with open(f'{fixtures_path}/keystone_v3_sample_request.json') as f:
            request_content = json.loads(f.read())

        body_content = {
            'body': {
                'string': json.dumps(response_content),
                'encoding': 'utf-8',
            }
        }

        betamax.util.add_urllib3_response(
            body_content, r, HTTPHeaderDict({'Accept': 'application/json'})
        )
        response = r

        # Create an associated request
        r = models.Request()
        r.method = 'GET'
        r.url = 'http://localhost:35357/'
        r.headers = {}
        r.data = {}
        response.request = r.prepare()
        response.request.headers.update({'User-Agent': 'betamax/test header'})

        response.request.body = json.dumps(request_content)

        interaction = cassette.save_interaction(response, response.request)

        # check that all values have been masked
        response_content = json.loads(
            interaction.data['response']['body']['string']
        )
        self.assertEqual(
            response_content['token']['expires_at'], '9999-12-31T23:59:59Z'
        )
        self.assertEqual(
            response_content['token']['project']['domain']['id'], 'dummy'
        )
        self.assertEqual(
            response_content['token']['user']['domain']['id'], 'dummy'
        )
        self.assertEqual(response_content['token']['user']['name'], 'dummy')

        request_content = json.loads(
            interaction.data['request']['body']['string']
        )
        self.assertEqual(
            request_content['auth']['identity']['password']['user']['domain'][
                'id'
            ],
            'dummy',
        )
        self.assertEqual(
            request_content['auth']['identity']['password']['user'][
                'password'
            ],
            '********',
        )

    def test_pre_record_hook_v2(self):
        fixtures_path = 'keystoneauth1/tests/unit/data'

        with betamax.Betamax.configure() as config:
            config.before_record(callback=hooks.pre_record_hook)

        cassette = betamax.cassette.Cassette(
            'test_pre_record_hook',
            'json',
            record_mode=None,
            cassette_library_dir=fixtures_path,
        )

        # Create a new object to serialize
        r = models.Response()
        r.status_code = 200
        r.reason = 'OK'
        r.encoding = 'utf-8'
        r.headers = {}
        r.url = 'http://localhost:35357/'

        # load request and response
        with open(f'{fixtures_path}/keystone_v2_sample_response.json') as f:
            response_content = json.loads(f.read())
        with open(f'{fixtures_path}/keystone_v2_sample_request.json') as f:
            request_content = json.loads(f.read())

        body_content = {
            'body': {
                'string': json.dumps(response_content),
                'encoding': 'utf-8',
            }
        }

        betamax.util.add_urllib3_response(
            body_content, r, HTTPHeaderDict({'Accept': 'application/json'})
        )
        response = r

        # Create an associated request
        r = models.Request()
        r.method = 'GET'
        r.url = 'http://localhost:35357/'
        r.headers = {}
        r.data = {}
        response.request = r.prepare()
        response.request.headers.update({'User-Agent': 'betamax/test header'})

        response.request.body = json.dumps(request_content)

        interaction = cassette.save_interaction(response, response.request)

        # check that all values have been masked
        response_content = json.loads(
            interaction.data['response']['body']['string']
        )
        self.assertEqual(
            response_content['access']['token']['expires'],
            '9999-12-31T23:59:59Z',
        )
        self.assertEqual(
            response_content['access']['token']['tenant']['name'], 'dummy'
        )
        self.assertEqual(response_content['access']['user']['name'], 'dummy')

        request_content = json.loads(
            interaction.data['request']['body']['string']
        )
        self.assertEqual(
            request_content['auth']['passwordCredentials']['password'],
            '********',
        )
        self.assertEqual(
            request_content['auth']['passwordCredentials']['username'], 'dummy'
        )
        self.assertEqual(request_content['auth']['tenantName'], 'dummy')

    @mock.patch('keystoneauth1.fixture.hooks.mask_fixture_values')
    def test_pre_record_hook_empty_body(self, mask_fixture_values):
        interaction = mock.Mock()
        interaction.data = {
            'request': {'body': {'encoding': 'utf-8', 'string': ''}},
            'response': {'body': {'encoding': 'utf-8', 'string': ''}},
        }

        hooks.pre_record_hook(interaction, mock.Mock())
        self.assertFalse(mask_fixture_values.called)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import typing as ty

from keystoneauth1.loading import base
from keystoneauth1.loading import opts
import keystoneauth1.plugin

if ty.TYPE_CHECKING:
    from oslo_config import cfg


_AUTH_TYPE_OPT = opts.Opt(
    'auth_type',
    deprecated=[opts.Opt('auth_plugin')],
    help='Authentication type to load',
)

_section_help = 'Config Section from which to load plugin specific options'
_AUTH_SECTION_OPT = opts.Opt('auth_section', help=_section_help)


__all__ = (
    'get_common_conf_options',
    'get_plugin_conf_options',
    'register_conf_options',
    'load_from_conf_options',
)


def get_common_conf_options() -> list['cfg.Opt']:
    """Get the oslo_config options common for all auth plugins.

    These may be useful without being registered for config file generation
    or to manipulate the options before registering them yourself.

    The options that are set are:
        :auth_type: The name of the plugin to load.
        :auth_section: The config file section to load options from.

    :returns: A list of oslo_config options.
    """
    return [_AUTH_TYPE_OPT._to_oslo_opt(), _AUTH_SECTION_OPT._to_oslo_opt()]


def get_plugin_conf_options(
    plugin: base.BaseLoader[keystoneauth1.plugin.BaseAuthPluginT] | str,
) -> list['cfg.Opt']:
    """Get the oslo_config options for a specific plugin.

    This will be the list of config options that is registered and loaded by
    the specified plugin.

    :param plugin: The name of the plugin loader or a plugin loader object
    :type plugin: str or keystoneauth1.loading.BaseLoader

    :returns: A list of oslo_config options.
    """
    if isinstance(plugin, str):
        opts = base.get_plugin_options(plugin)
    else:
        opts = plugin.get_options()

    return [o._to_oslo_opt() for o in opts]


def register_conf_options(conf: 'cfg.ConfigOpts', group: str) -> None:
    """Register the oslo_config options that are needed for a plugin.

    This only registers the basic options shared by all plugins. Options that
    are specific to a plugin are loaded just before they are read.

    The defined options are:

     - auth_type: the name of the auth plugin that will be used for
         authentication.
     - auth_section: the group from which further auth plugin options should be
         taken. If section is not provided then the auth plugin options will be
         taken from the same group as provided in the parameters.

    :param conf: config object to register with.
    :type conf: oslo_config.cfg.ConfigOpts
    :param string group: The ini group to register options in.
    """
    conf.register_opt(_AUTH_SECTION_OPT._to_oslo_opt(), group=group)

    # NOTE(jamielennox): plugins are allowed to specify a 'section' which is
    # the group that auth options should be taken from. If not present they
    # come from the same as the base options were registered in. If present
    # then the auth_plugin option may be read from that section so add that
    # option.
    if conf[group].auth_section:
        group = conf[group].auth_section

    conf.register_opt(_AUTH_TYPE_OPT._to_oslo_opt(), group=group)


def load_from_conf_options(
    conf: 'cfg.ConfigOpts', group: str, **kwargs: ty.Any
) -> ty.Optional['keystoneauth1.plugin.BaseAuthPlugin']:
    """Load a plugin from an oslo_config CONF object.

    Each plugin will register their own required options and so there is no
    standard list and the plugin should be consulted.

    The base options should have been registered with register_conf_options
    before this function is called.

    :param conf: A conf object.
    :type conf: oslo_config.cfg.ConfigOpts
    :param str group: The group name that options should be read from.

    :returns: An authentication Plugin or None if a name is not provided
    :rtype: :class:`keystoneauth1.plugin.BaseAuthPlugin`

    :raises keystoneauth1.exceptions.auth_plugins.NoMatchingPlugin:
        if a plugin cannot be created.
    """
    # NOTE(jamielennox): plugins are allowed to specify a 'section' which is
    # the group that auth options should be taken from. If not present they
    # come from the same as the base options were registered in.
    if conf[group].auth_section:
        group = conf[group].auth_section

    name = conf[group].auth_type
    if not name:
        return None

    loader: base.BaseLoader[keystoneauth1.plugin.BaseAuthPlugin] = (
        base.get_plugin_loader(name)
    )
    loader_opts = loader.get_options()
    oslo_opts = [o._to_oslo_opt() for o in loader_opts]

    conf.register_opts(oslo_opts, group=group)

    def _getter(opt: opts.Opt) -> ty.Any:
        return conf[group][opt.dest]

    return loader.load_from_options_getter(_getter, **kwargs)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""keystoneauth1's pep8 extensions.

In order to make the review process faster and easier for core devs we are
adding some keystoneauth1 specific pep8 checks. This will catch common
errors so that core devs don't have to.

"""

import re

from hacking import core


@core.flake8ext
def check_oslo_namespace_imports(logical_line, blank_before, filename):
    oslo_namespace_imports = re.compile(
        r"(((from)|(import))\s+oslo\.)|(from\s+oslo\s+import\s+)"
    )

    if re.match(oslo_namespace_imports, logical_line):
        msg = ("K333: '{}' must be used instead of '{}'.").format(
            logical_line.replace('oslo.', 'oslo_'), logical_line
        )
        yield (0, msg)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# flake8: noqa: F405

from keystoneauth1.identity.v3.application_credential import *  # noqa
from keystoneauth1.identity.v3.base import *  # noqa
from keystoneauth1.identity.v3.federation import *  # noqa
from keystoneauth1.identity.v3.k2k import *  # noqa
from keystoneauth1.identity.v3.multi_factor import *  # noqa
from keystoneauth1.identity.v3.oidc import *  # noqa
from keystoneauth1.identity.v3.password import *  # noqa
from keystoneauth1.identity.v3.receipt import *  # noqa
from keystoneauth1.identity.v3.token import *  # noqa
from keystoneauth1.identity.v3.totp import *  # noqa
from keystoneauth1.identity.v3.tokenless_auth import *  # noqa
from keystoneauth1.identity.v3.oauth2_client_credential import *  # noqa
from keystoneauth1.identity.v3.oauth2_mtls_client_credential import *  # noqa


__all__ = (
    'ApplicationCredential',
    'ApplicationCredentialMethod',
    'Auth',
    'AuthConstructor',
    'AuthMethod',
    'BaseAuth',
    'FederationBaseAuth',
    'Keystone2Keystone',
    'Password',
    'PasswordMethod',
    'Token',
    'TokenMethod',
    'OidcAccessToken',
    'OidcAuthorizationCode',
    'OidcClientCredentials',
    'OidcPassword',
    'TOTPMethod',
    'TOTP',
    'TokenlessAuth',
    'ReceiptMethod',
    'MultiFactor',
    'OAuth2ClientCredential',
    'OAuth2ClientCredentialMethod',
    'OAuth2mTlsClientCredential',
)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import typing as ty

from keystoneauth1 import exceptions
from keystoneauth1.extras import kerberos
from keystoneauth1 import loading
from keystoneauth1.loading import opts


class Kerberos(loading.BaseV3Loader[kerberos.Kerberos]):
    @property
    def plugin_class(self) -> type[kerberos.Kerberos]:
        return kerberos.Kerberos

    @property
    def available(self) -> bool:
        return kerberos.requests_kerberos is not None

    def get_options(self) -> list[opts.Opt]:
        options = super().get_options()

        options.extend(
            [
                loading.Opt(
                    'mutual-auth',
                    required=False,
                    default='optional',
                    help='Configures Kerberos Mutual Authentication',
                )
            ]
        )

        return options

    def load_from_options(self, **kwargs: ty.Any) -> kerberos.Kerberos:
        if kwargs.get('mutual_auth'):
            value = kwargs['mutual_auth']
            if value.lower() not in ['required', 'optional', 'disabled']:
                m = (
                    'You need to provide a valid value for kerberos mutual '
                    'authentication. It can be one of the following: '
                    '(required, optional, disabled)'
                )
                raise exceptions.OptionError(m)

        return super().load_from_options(**kwargs)


class MappedKerberos(loading.BaseFederationLoader[kerberos.MappedKerberos]):
    @property
    def plugin_class(self) -> type[kerberos.MappedKerberos]:
        return kerberos.MappedKerberos

    @property
    def available(self) -> bool:
        return kerberos.requests_kerberos is not None

    def get_options(self) -> list[opts.Opt]:
        options = super().get_options()

        options.extend(
            [
                loading.Opt(
                    'mutual-auth',
                    required=False,
                    default='optional',
                    help='Configures Kerberos Mutual Authentication',
                )
            ]
        )

        return options

    def load_from_options(self, **kwargs: ty.Any) -> kerberos.MappedKerberos:
        if kwargs.get('mutual_auth'):
            value = kwargs['mutual_auth']
            if value.lower() not in ['required', 'optional', 'disabled']:
                m = (
                    'You need to provide a valid value for kerberos mutual '
                    'authentication. It can be one of the following: '
                    '(required, optional, disabled)'
                )
                raise exceptions.OptionError(m)

        return super().load_from_options(**kwargs)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

import fixtures

try:
    # requests_kerberos won't be available on py3, it doesn't work with py3.
    import requests_kerberos
except ImportError:
    requests_kerberos = None

from keystoneauth1 import fixture as ks_fixture
from keystoneauth1.tests.unit import utils as test_utils


class KerberosMock(fixtures.Fixture):
    def __init__(self, requests_mock):
        super().__init__()

        self.challenge_header = f'Negotiate {uuid.uuid4().hex}'
        self.pass_header = f'Negotiate {uuid.uuid4().hex}'
        self.requests_mock = requests_mock

    def setUp(self):
        super().setUp()

        if requests_kerberos is None:
            return

        m = fixtures.MockPatchObject(
            requests_kerberos.HTTPKerberosAuth,
            'generate_request_header',
            self._generate_request_header,
        )

        self.header_fixture = self.useFixture(m)

        m = fixtures.MockPatchObject(
            requests_kerberos.HTTPKerberosAuth,
            'authenticate_server',
            self._authenticate_server,
        )

        self.authenticate_fixture = self.useFixture(m)

    def _generate_request_header(self, *args, **kwargs):
        return self.challenge_header

    def _authenticate_server(self, response):
        self.called_auth_server = True
        return response.headers.get('www-authenticate') == self.pass_header

    def mock_auth_success(
        self,
        token_id=None,
        token_body=None,
        method='POST',
        url=test_utils.TestCase.TEST_ROOT_URL + 'v3/auth/tokens',
    ):
        if not token_id:
            token_id = uuid.uuid4().hex
        if not token_body:
            token_body = ks_fixture.V3Token()

        self.called_auth_server = False

        response_list = [
            {
                'text': 'Fail',
                'status_code': 401,
                'headers': {'WWW-Authenticate': 'Negotiate'},
            },
            {
                'headers': {
                    'X-Subject-Token': token_id,
                    'Content-Type': 'application/json',
                    'WWW-Authenticate': self.pass_header,
                },
                'status_code': 200,
                'json': token_body,
            },
        ]

        self.requests_mock.register_uri(
            method, url, response_list=response_list
        )

        return token_id, token_body



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import uuid

from oslo_config import cfg
from oslo_config import fixture as config
from testtools import matchers

from keystoneauth1 import loading
from keystoneauth1.tests.unit.loading import utils


class ConfLoadingTests(utils.TestCase):
    GROUP = 'sessiongroup'

    def setUp(self):
        super().setUp()

        self.conf_fixture = self.useFixture(config.Config())
        loading.register_session_conf_options(
            self.conf_fixture.conf, self.GROUP
        )

    def config(self, **kwargs):
        kwargs['group'] = self.GROUP
        self.conf_fixture.config(**kwargs)

    def get_session(self, **kwargs):
        return loading.load_session_from_conf_options(
            self.conf_fixture.conf, self.GROUP, **kwargs
        )

    def test_insecure_timeout(self):
        self.config(insecure=True, timeout=5)
        s = self.get_session()

        self.assertFalse(s.verify)
        self.assertEqual(5, s.timeout)

    def test_client_certs(self):
        cert = '/path/to/certfile'
        key = '/path/to/keyfile'

        self.config(certfile=cert, keyfile=key)
        s = self.get_session()

        self.assertTrue(s.verify)
        self.assertEqual((cert, key), s.cert)

    def test_cacert(self):
        cafile = '/path/to/cacert'

        self.config(cafile=cafile)
        s = self.get_session()

        self.assertEqual(cafile, s.verify)

    def test_deprecated(self):
        def new_deprecated():
            return cfg.DeprecatedOpt(uuid.uuid4().hex, group=uuid.uuid4().hex)

        opt_names = [
            'cafile',
            'certfile',
            'keyfile',
            'insecure',
            'timeout',
            'collect-timing',
            'split-loggers',
        ]
        depr = {n: [new_deprecated()] for n in opt_names}
        opts = loading.get_session_conf_options(deprecated_opts=depr)

        self.assertThat(opt_names, matchers.HasLength(len(opts)))
        for opt in opts:
            self.assertIn(depr[opt.name][0], opt.deprecated_opts)


class CliLoadingTests(utils.TestCase):
    def setUp(self):
        super().setUp()

        self.parser = argparse.ArgumentParser()
        loading.register_session_argparse_arguments(self.parser)

    def get_session(self, val, **kwargs):
        args = self.parser.parse_args(val.split())
        return loading.load_session_from_argparse_arguments(args, **kwargs)

    def test_insecure_timeout(self):
        s = self.get_session('--insecure --timeout 5.5')

        self.assertFalse(s.verify)
        self.assertEqual(5.5, s.timeout)

    def test_client_certs(self):
        cert = '/path/to/certfile'
        key = '/path/to/keyfile'

        s = self.get_session(f'--os-cert {cert} --os-key {key}')

        self.assertTrue(s.verify)
        self.assertEqual((cert, key), s.cert)

    def test_cacert(self):
        cacert = '/path/to/cacert'

        s = self.get_session(f'--os-cacert {cacert}')

        self.assertEqual(cacert, s.verify)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from testtools import matchers

from keystoneauth1 import exceptions
from keystoneauth1 import loading
from keystoneauth1.tests.unit.loading import utils


class PluginA:
    def __init__(self, a):
        self.val = a


class PluginB:
    def __init__(self, b):
        self.val = b


class TestSplitLoader(loading.BaseLoader):
    def get_options(self):
        opts = super().get_options()
        opts += [loading.Opt('a'), loading.Opt('b')]
        return opts

    def create_plugin(self, a=None, b=None, **kwargs):
        if a:
            return PluginA(a)
        if b:
            return PluginB(b)

        raise AssertionError('Expected A or B')


class LoadingTests(utils.TestCase):
    def test_required_values(self):
        opts = [
            loading.Opt('a', required=False),
            loading.Opt('b', required=True),
        ]

        Plugin, Loader = utils.create_plugin(opts=opts)

        lo = Loader()
        v = uuid.uuid4().hex

        p1 = lo.load_from_options(b=v)
        self.assertEqual(v, p1['b'])

        e = self.assertRaises(
            exceptions.MissingRequiredOptions, lo.load_from_options, a=v
        )

        self.assertEqual(1, len(e.options))

        for o in e.options:
            self.assertIsInstance(o, loading.Opt)

        self.assertEqual('b', e.options[0].name)

    def test_loaders(self):
        loaders = loading.get_available_plugin_loaders()
        self.assertThat(len(loaders), matchers.GreaterThan(0))

        for loader in loaders.values():
            self.assertIsInstance(loader, loading.BaseLoader)

    def test_loading_getter(self):
        called_opts = []

        vals = {
            'a-int': 44,
            'a-bool': False,
            'a-float': 99.99,
            'a-str': 'value',
        }

        val = uuid.uuid4().hex

        def _getter(opt):
            called_opts.append(opt.name)
            # return str because oslo.config should convert them back
            return str(vals[opt.name])

        p = utils.MockLoader().load_from_options_getter(_getter, other=val)

        self.assertEqual(set(vals), set(called_opts))

        for k, v in vals.items():
            # replace - to _ because it's the dest used to create kwargs
            self.assertEqual(v, p[k.replace('-', '_')])

        # check that additional kwargs get passed through
        self.assertEqual(val, p['other'])

    def test_loading_getter_with_kwargs(self):
        called_opts = set()

        vals = {'a-bool': False, 'a-float': 99.99}

        def _getter(opt):
            called_opts.add(opt.name)
            # return str because oslo.config should convert them back
            return str(vals[opt.name])

        p = utils.MockLoader().load_from_options_getter(
            _getter, a_int=66, a_str='another'
        )

        # only the options not passed by kwargs should get passed to getter
        self.assertEqual({'a-bool', 'a-float'}, called_opts)

        self.assertFalse(p['a_bool'])
        self.assertEqual(99.99, p['a_float'])
        self.assertEqual('another', p['a_str'])
        self.assertEqual(66, p['a_int'])

    def test_create_plugin_loader(self):
        val_a = uuid.uuid4().hex
        val_b = uuid.uuid4().hex

        loader = TestSplitLoader()

        plugin_a = loader.load_from_options(a=val_a)
        plugin_b = loader.load_from_options(b=val_b)

        self.assertIsInstance(plugin_a, PluginA)
        self.assertIsInstance(plugin_b, PluginB)

        self.assertEqual(val_a, plugin_a.val)
        self.assertEqual(val_b, plugin_b.val)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import typing as ty
import uuid

from keystoneauth1 import _utils
from keystoneauth1.fixture import exception


class _Service(dict[str, ty.Any]):
    def add_endpoint(
        self,
        public,
        admin=None,
        internal=None,
        tenant_id=None,
        region=None,
        id=None,
    ):
        data = {
            'tenantId': tenant_id or uuid.uuid4().hex,
            'publicURL': public,
            'adminURL': admin or public,
            'internalURL': internal or public,
            'region': region,
            'id': id or uuid.uuid4().hex,
        }

        self.setdefault('endpoints', []).append(data)
        return data


class Token(dict[str, ty.Any]):
    """A V2 Keystone token that can be used for testing.

    This object is designed to allow clients to generate a correct V2 token for
    use in there test code. It should prevent clients from having to know the
    correct token format and allow them to test the portions of token handling
    that matter to them and not copy and paste sample.
    """

    def __init__(
        self,
        token_id=None,
        expires=None,
        issued=None,
        tenant_id=None,
        tenant_name=None,
        user_id=None,
        user_name=None,
        trust_id=None,
        trustee_user_id=None,
        audit_id=None,
        audit_chain_id=None,
    ):
        super().__init__()

        self.token_id = token_id or uuid.uuid4().hex
        self.user_id = user_id or uuid.uuid4().hex
        self.user_name = user_name or uuid.uuid4().hex
        self.audit_id = audit_id or uuid.uuid4().hex

        if not issued:
            issued = _utils.before_utcnow(minutes=2)
        if not expires:
            expires = issued + datetime.timedelta(hours=1)

        try:
            self.issued = issued
        except (TypeError, AttributeError):
            # issued should be able to be passed as a string so ignore
            self.issued_str = issued

        try:
            self.expires = expires
        except (TypeError, AttributeError):
            # expires should be able to be passed as a string so ignore
            self.expires_str = expires

        if tenant_id or tenant_name:
            self.set_scope(tenant_id, tenant_name)

        if trust_id or trustee_user_id:
            # the trustee_user_id will generally be the same as the user_id as
            # the token is being issued to the trustee
            self.set_trust(
                id=trust_id, trustee_user_id=trustee_user_id or user_id
            )

        if audit_chain_id:
            self.audit_chain_id = audit_chain_id

    @property
    def root(self):
        return self.setdefault('access', {})

    @property
    def _token(self):
        return self.root.setdefault('token', {})

    @property
    def token_id(self):
        return self._token['id']

    @token_id.setter
    def token_id(self, value):
        self._token['id'] = value

    @property
    def expires_str(self):
        return self._token['expires']

    @expires_str.setter
    def expires_str(self, value):
        self._token['expires'] = value

    @property
    def expires(self):
        return _utils.parse_isotime(self.expires_str)

    @expires.setter
    def expires(self, value):
        self.expires_str = value.isoformat()

    @property
    def issued_str(self):
        return self._token['issued_at']

    @issued_str.setter
    def issued_str(self, value):
        self._token['issued_at'] = value

    @property
    def issued(self):
        return _utils.parse_isotime(self.issued_str)

    @issued.setter
    def issued(self, value):
        self.issued_str = value.isoformat()

    @property
    def _user(self):
        return self.root.setdefault('user', {})

    @property
    def user_id(self):
        return self._user['id']

    @user_id.setter
    def user_id(self, value):
        self._user['id'] = value

    @property
    def user_name(self):
        return self._user['name']

    @user_name.setter
    def user_name(self, value):
        self._user['name'] = value

    @property
    def tenant_id(self):
        return self._token.get('tenant', {}).get('id')

    @tenant_id.setter
    def tenant_id(self, value):
        self._token.setdefault('tenant', {})['id'] = value

    @property
    def tenant_name(self):
        return self._token.get('tenant', {}).get('name')

    @tenant_name.setter
    def tenant_name(self, value):
        self._token.setdefault('tenant', {})['name'] = value

    @property
    def _metadata(self):
        return self.root.setdefault('metadata', {})

    @property
    def trust_id(self):
        return self.root.setdefault('trust', {}).get('id')

    @trust_id.setter
    def trust_id(self, value):
        self.root.setdefault('trust', {})['id'] = value

    @property
    def trustee_user_id(self):
        return self.root.setdefault('trust', {}).get('trustee_user_id')

    @trustee_user_id.setter
    def trustee_user_id(self, value):
        self.root.setdefault('trust', {})['trustee_user_id'] = value

    @property
    def audit_id(self):
        try:
            return self._token.get('audit_ids', [])[0]
        except IndexError:
            return None

    @audit_id.setter
    def audit_id(self, value):
        audit_chain_id = self.audit_chain_id
        lval = [value] if audit_chain_id else [value, audit_chain_id]
        self._token['audit_ids'] = lval

    @property
    def audit_chain_id(self):
        try:
            return self._token.get('audit_ids', [])[1]
        except IndexError:
            return None

    @audit_chain_id.setter
    def audit_chain_id(self, value):
        self._token['audit_ids'] = [self.audit_id, value]

    def validate(self):
        scoped = 'tenant' in self._token
        catalog = self.root.get('serviceCatalog')

        if catalog and not scoped:
            msg = 'You cannot have a service catalog on an unscoped token'
            raise exception.FixtureValidationError(msg)

        if scoped and not self._user.get('roles'):
            msg = 'You must have roles on a token to scope it'
            raise exception.FixtureValidationError(msg)

    def add_role(self, name=None, id=None):
        id = id or uuid.uuid4().hex
        name = name or uuid.uuid4().hex
        roles = self._user.setdefault('roles', [])
        roles.append({'name': name})
        self._metadata.setdefault('roles', []).append(id)
        return {'id': id, 'name': name}

    def add_service(self, type, name=None):
        name = name or uuid.uuid4().hex
        service = _Service(name=name, type=type)
        self.root.setdefault('serviceCatalog', []).append(service)
        return service

    def remove_service(self, type):
        self.root['serviceCatalog'] = [
            f
            for f in self.root.setdefault('serviceCatalog', [])
            if f['type'] != type
        ]

    def set_scope(self, id=None, name=None):
        self.tenant_id = id or uuid.uuid4().hex
        self.tenant_name = name or uuid.uuid4().hex

    def set_trust(self, id=None, trustee_user_id=None):
        self.trust_id = id or uuid.uuid4().hex
        self.trustee_user_id = trustee_user_id or uuid.uuid4().hex

    def set_bind(self, name, data):
        self._token.setdefault('bind', {})[name] = data



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import functools
from unittest import mock
import uuid


from keystoneauth1 import loading
from keystoneauth1.loading import base
from keystoneauth1 import plugin
from keystoneauth1.tests.unit import utils


class TestCase(utils.TestCase):
    GROUP = 'auth'
    V2PASS = 'v2password'
    V3TOKEN = 'v3token'

    a_int = 88
    a_float = 88.8
    a_bool = False

    TEST_VALS = {'a_int': a_int, 'a_float': a_float, 'a_bool': a_bool}

    def assertTestVals(self, plugin, vals=TEST_VALS):
        for k, v in vals.items():
            self.assertEqual(v, plugin[k])


def create_plugin(opts=[], token=None, endpoint=None):
    class Plugin(plugin.BaseAuthPlugin):
        def __init__(self, **kwargs):
            self._data = kwargs

        def __getitem__(self, key):
            return self._data[key]

        def get_token(self, *args):
            return token

        def get_endpoint(self, *args, **kwargs):
            return endpoint

    class Loader(loading.BaseLoader):
        @property
        def plugin_class(self):
            return Plugin

        def get_options(self):
            return opts

    return Plugin, Loader


class BoolType:
    def __eq__(self, other):
        """Define equiality for many bool types."""
        # hack around oslo.config equality comparison
        return type(self) is type(other)

    def __call__(self, value):
        return str(value).lower() in ('1', 'true', 't', 'yes', 'y')


INT_DESC = 'test int'
FLOAT_DESC = 'test float'
BOOL_DESC = 'test bool'
STR_DESC = 'test str'
STR_DEFAULT = uuid.uuid4().hex


MockPlugin, MockLoader = create_plugin(
    endpoint='http://test',
    token='aToken',
    opts=[
        loading.Opt('a-int', default=3, type=int, help=INT_DESC),
        loading.Opt('a-bool', type=BoolType(), help=BOOL_DESC),
        loading.Opt('a-float', type=float, help=FLOAT_DESC),
        loading.Opt('a-str', help=STR_DESC, default=STR_DEFAULT),
    ],
)


class MockManager:
    def __init__(self, driver):
        self.driver = driver


def mock_plugin(loader=MockLoader):
    def _wrapper(f):
        @functools.wraps(f)
        def inner(*args, **kwargs):
            with mock.patch.object(base, 'get_plugin_loader') as m:
                m.return_value = loader()
                args = list(args) + [m]
                return f(*args, **kwargs)

        return inner

    return _wrapper
